import asyncio
import base64
import os
import pathlib
import shutil
from datetime import datetime
from typing import Any, AsyncGenerator, Dict, List, Optional
from unittest.mock import call

# from fastapi.testclient import TestClient
import pytest  # type: ignore
from async_asgi_testclient import TestClient  # type: ignore
from pytest_mock import MockFixture  # type: ignore

from jupyter_d1 import app
from jupyter_d1.settings import settings

from .utils import (
    collect_websocket_messages,
    filter_websocket_collection,
    get_superuser_token,
    receive_json,
    wait_for_event,
    workdir_1,
)

# All test coroutines will be treated as marked.
# pytestmark = pytest.mark.asyncio

WEBSOCKET_SLEEP_TIME = 10.1


@pytest.fixture()
async def client() -> AsyncGenerator:
    async with TestClient(app) as c:
        yield c


@pytest.fixture()
async def superuser_token_headers(client: TestClient) -> Dict[str, str]:
    r = await client.get(
        "/login/access-token", headers={"Authorization": "test9token_4"}
    )
    token = r.json()["token"]["access_token"]
    headers = {"Authorization": f"Bearer {token}"}
    return headers


@pytest.fixture(scope="function")
async def clear_notebooks(
    client: TestClient, superuser_token_headers: Dict[str, str]
) -> AsyncGenerator:
    yield
    response = await client.delete(
        "/notebooks", headers=superuser_token_headers
    )
    assert response.status_code == 204
    response = await client.get("/notebooks", headers=superuser_token_headers)
    assert response.status_code == 200


async def upload_notebook(
    client: TestClient,
    token_headers: Dict[str, str],
    filename: str = "simple.ipynb",
) -> str:
    nb_filename = f"jupyter_d1/tests/notebooks/{filename}"
    nb_json = open(nb_filename).read()
    response = await client.post(
        "/notebooks/upload",
        query_string={"filename": filename},
        data=nb_json,
        headers=token_headers,
    )
    assert response.status_code == 201
    path = response.json()["path"]

    response = await client.get(
        f"/notebooks/open/?filepath={path}", headers=token_headers
    )
    assert response.status_code == 201
    resp_json = response.json()["notebook"]
    uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

    # Give kernel time to fully start, kernel messages can get
    # dropped otherwise
    await asyncio.sleep(2)
    return uuid


def assert_cell_positions(cells: List[Dict[str, Any]]):
    assert [c["metadata"]["jupyter_d1"]["position"] for c in cells] == list(
        range(0, len(cells))
    )


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestNotebookWebSocket:
    @pytest.mark.asyncio
    async def test_notebook_cell_execute(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:

            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            msg_id = response.json()["kernel_message"]["message_id"]

            # first two chunks received should be a execution count update and
            # cell_update with an execution_state of 'busy' when processing
            # starts. Order is variable, so here we handle either order.
            def assert_json_content(data, execution_count=None):
                if "cell_update" in data:
                    cell_data = data["cell_update"][0]
                    d1_metadata = cell_data["metadata"]["jupyter_d1"]
                    assert d1_metadata["execution_state"] == "busy"
                    assert d1_metadata["uuid"] == cell_uuid
                    assert cell_data["execution_count"] == execution_count
                    return None
                elif "cell_execution_reply" in data:
                    assert data["cell_execution_reply"]["execution_count"] == 1
                    assert data["cell_execution_reply"]["cell_id"] == cell_uuid
                    assert data["cell_execution_reply"]["parent_id"] == msg_id
                    return 1
                else:
                    raise AssertionError(f"unexpected cell type: {data}")

            data = await receive_json(
                websocket, msg_types=["cell_update", "cell_execution_reply"]
            )
            exec_count = assert_json_content(data)
            data = await receive_json(
                websocket, msg_types=["cell_update", "cell_execution_reply"]
            )
            assert_json_content(data, exec_count)

            # third chunk received should be a cell_update with
            # the new output, and an execution_state of 'busy'
            # since it is still processing
            data = await receive_json(websocket, msg_types=["cell_update"])
            expected = {
                "cell_update": [
                    {
                        "cell_type": "code",
                        "execution_count": 1,
                        "metadata": {
                            "jupyter_d1": {
                                "position": 2,
                                "uuid": cell_uuid,
                                "notebook_uuid": uuid,
                                "execution_state": "busy",
                            }
                        },
                        "outputs": [
                            {
                                "data": {"text/plain": "7"},
                                "execution_count": 1,
                                "metadata": {},
                                "output_type": "execute_result",
                            }
                        ],
                        "source": "2+5",
                    }
                ]
            }

            # from pprint import pprint
            # print("data")
            # pprint(data)
            # print("expected")
            # pprint(expected)
            assert data == expected

            # third chunk received should be a cell_update with
            # an execution_state of 'idle' becuase it's all done
            data = await receive_json(websocket, msg_types=["cell_update"])
            d1_metadata = data["cell_update"][0]["metadata"]["jupyter_d1"]
            assert d1_metadata["execution_state"] == "idle"

            data = await receive_json(websocket, msg_types=["vars"])
            assert len(data["vars"]) == 14

    @pytest.mark.asyncio
    async def test_notebook_sequential_execute(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell_uuid1 = cells[1]["metadata"]["jupyter_d1"]["uuid"]
        cell_uuid2 = cells[2]["metadata"]["jupyter_d1"]["uuid"]
        cell_uuid3 = cells[3]["metadata"]["jupyter_d1"]["uuid"]

        code1 = """import time; time.sleep(1.1); print("checkpoint1")"""
        code2 = """import time; time.sleep(1.2); print("checkpoint2")"""
        code3 = """import time; time.sleep(1.3); print("checkpoint3")"""

        params1 = {"source": code1}
        params2 = {"source": code2}
        params3 = {"source": code3}

        resp1 = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid1}",
            json=params1,
            headers=superuser_token_headers,
        )
        assert resp1.status_code == 200

        resp2 = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid2}",
            json=params2,
            headers=superuser_token_headers,
        )
        assert resp2.status_code == 200

        resp3 = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid3}",
            json=params3,
            headers=superuser_token_headers,
        )
        assert resp3.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            pass

            # await asyncio.sleep(WEBSOCKET_SLEEP_TIME)
            await asyncio.sleep(3.5)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid1}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid2}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid3}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            # Messages on the web socket can arrive in a slightly different
            # order from run to run.  The 'cell_update' messages arrive arround
            # the same time as some of the cell_execution messages.  It doesn't
            # really matter, but it makes testing tricky.  For a stable test,
            # gather all the responses and split them into two lists, one for
            # cell_update messages and one for all the others.

            # The most important thing for this test is that all the
            # "cell_execution_request" messages arrive first as the cells are
            # queued for execution and the "cell_execution_input",
            # "cell_execution_reply", etc messages show up later as the cells
            # are processed.

            cell_update_msgs = []
            non_cell_update_msgs = []
            for i in range(19):
                data = await receive_json(websocket)
                if "cell_update" in data.keys():
                    cell_update_msgs.append(data)
                else:
                    non_cell_update_msgs.append(data)

            # should get three execution requests
            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_request" in data.keys()
            msg_id1 = data["cell_execution_request"]["message_id"]
            assert len(msg_id1) == 35

            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_request" in data.keys()
            msg_id2 = data["cell_execution_request"]["message_id"]
            assert len(msg_id2) == 35

            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_request" in data.keys()
            msg_id3 = data["cell_execution_request"]["message_id"]
            assert len(msg_id3) == 35

            # cell update for the first executing cell
            data = cell_update_msgs.pop(0)
            assert "cell_update" in data.keys()
            assert (
                data["cell_update"][0]["metadata"]["jupyter_d1"]["uuid"]
                == cell_uuid1
            )

            # cell_execution_input for the first cell's code being executed
            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_input" in data
            assert (
                data["cell_execution_input"]["content"]["code"] == code1
            ), f'Failed to find "{code1}" in cell_stream data: {data}'
            assert (
                data["cell_execution_input"]["parent_id"] == msg_id1
            ), f"Failed to find expected parent_id"

            # stream output for the first cell's execution
            data = non_cell_update_msgs.pop(0)
            assert "cell_stream" in data.keys()
            assert (
                data["cell_stream"]["output"]["text"] == "checkpoint1\n"
            ), f"Failed to find checkpoint1 in cell_stream data: {data}"

            # execution_reply for the end of the first cell's execution
            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_reply" in data.keys()
            assert data["cell_execution_reply"]["parent_id"] == msg_id1

            # cell update with the results of the first cell's execution
            data = cell_update_msgs.pop(0)
            assert "cell_update" in data.keys()
            assert (
                data["cell_update"][0]["metadata"]["jupyter_d1"]["uuid"]
                == cell_uuid1
            )

            # cell update for the begining of the second cell's execution
            data = cell_update_msgs.pop(0)
            assert "cell_update" in data.keys()
            assert (
                data["cell_update"][0]["metadata"]["jupyter_d1"]["uuid"]
                == cell_uuid2
            )

            # cell_input update with the second cell's code being executed
            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_input" in data
            assert (
                data["cell_execution_input"]["content"]["code"] == code2
            ), f'Failed to find "{code2}" in cell_stream data: {data}'
            assert (
                data["cell_execution_input"]["parent_id"] == msg_id2
            ), f"Failed to find expected parent_id"

            # stream output of the second cell's execution
            data = non_cell_update_msgs.pop(0)
            assert "cell_stream" in data.keys()
            assert (
                data["cell_stream"]["output"]["text"] == "checkpoint2\n"
            ), f"Failed to find checkpoint2 in cell_stream data: {data}"

            # exectuion_reply for the end of the second cell's execution
            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_reply" in data.keys()
            assert data["cell_execution_reply"]["parent_id"] == msg_id2

            # cell_update with the results of the second cell
            data = cell_update_msgs.pop(0)
            assert "cell_update" in data.keys()
            assert (
                data["cell_update"][0]["metadata"]["jupyter_d1"]["uuid"]
                == cell_uuid2
            )

            # cell_update as the third cell begins execution
            data = cell_update_msgs.pop(0)
            assert "cell_update" in data.keys()
            assert (
                data["cell_update"][0]["metadata"]["jupyter_d1"]["uuid"]
                == cell_uuid3
            )

            # cell_execution_input with third cell code about to be executed
            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_input" in data
            assert (
                data["cell_execution_input"]["content"]["code"] == code3
            ), f'Failed to find "{code3}" in cell_stream data: {data}'
            assert (
                data["cell_execution_input"]["parent_id"] == msg_id3
            ), f"Failed to find expected parent_id"

            # stream output of the third cell
            data = non_cell_update_msgs.pop(0)
            assert "cell_stream" in data.keys()
            assert (
                data["cell_stream"]["output"]["text"] == "checkpoint3\n"
            ), f"Failed to find checkpoint3 in cell_stream data: {data}"

            # execution reply for finishing the third cell's execution
            data = non_cell_update_msgs.pop(0)
            assert "cell_execution_reply" in data.keys()
            assert data["cell_execution_reply"]["parent_id"] == msg_id3

            # cell_update for cell three with the output
            data = cell_update_msgs.pop(0)
            assert "cell_update" in data.keys()
            assert (
                data["cell_update"][0]["metadata"]["jupyter_d1"]["uuid"]
                == cell_uuid3
            )

            # var update
            data = non_cell_update_msgs.pop(0)
            assert "vars" in data.keys()

    @pytest.mark.asyncio
    async def test_cell_patch_and_execute(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]

        params = {"source": "7-9"}
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        assert len(response.json()["vars"]) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.patch(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                json=params,
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            expected = {
                "cell_update": [
                    {
                        "cell_type": "code",
                        "execution_count": 1,
                        "metadata": {
                            "jupyter_d1": {
                                "position": 1,
                                "uuid": cell_uuid,
                                "notebook_uuid": uuid,
                                "execution_state": "busy",
                            }
                        },
                        "outputs": [
                            {
                                "data": {"text/plain": "-2"},
                                "execution_count": 1,
                                "metadata": {},
                                "output_type": "execute_result",
                            }
                        ],
                        "source": "7-9",
                    }
                ]
            }

            patch_json = response.json()
            # response should contain both the 'cell','kernel_message' wrappers
            assert "cell" in patch_json.keys()
            msg_id = response.json()["kernel_message"]["message_id"]
            assert len(msg_id) == 35

            exec_reply = await receive_json(
                websocket, msg_types=["cell_execution_reply"]
            )
            assert exec_reply["cell_execution_reply"]["cell_id"] == cell_uuid
            assert exec_reply["cell_execution_reply"]["parent_id"] == msg_id

            data = await receive_json(websocket, msg_types=["cell_update"])
            assert data == expected

    @pytest.mark.asyncio
    async def test_notebook_cell_add(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)
            params = {
                "cell_type": "markdown",
                "source": "__**Hello Callisto!**__",
            }
            response = await client.post(
                f"/notebooks/{uuid}/cells",
                json=params,
                headers=superuser_token_headers,
            )
            assert response.status_code == 201

            data = await receive_json(websocket, msg_types=["cell_add"])
            cells = data["cell_add"]
            assert len(cells) == 5  # should dump all the cells
            cell = cells[-1]
            cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

            expected = {
                "cell_type": "markdown",
                "metadata": {
                    "jupyter_d1": {
                        "position": 4,
                        "uuid": cell_uuid,
                        "notebook_uuid": uuid,
                    }
                },
                "source": "__**Hello Callisto!**__",
            }

            # from pprint import pprint
            # print("data")
            # pprint(data)
            # print("expected")
            # pprint(expected)
            assert cell == expected

    @pytest.mark.asyncio
    async def test_notebook_cell_patch(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            # get an existing cell
            response = await client.get(
                f"/notebooks/{uuid}/cells", headers=superuser_token_headers
            )
            assert response.status_code == 200
            cells = response.json()["cells"]
            assert cells[1]["source"] == 'print("Larry the Llama")'

            cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]
            params = {"source": "__Banana Rama!__"}
            response = await client.patch(
                f"/notebooks/{uuid}/cells/{cell_uuid}",
                json=params,
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            data = await receive_json(websocket)
            cells = data["cell_update"]
            assert len(cells) == 1
            cell = cells[0]

            expected = {
                "cell_type": "code",
                "execution_count": None,
                "metadata": {
                    "jupyter_d1": {
                        "execution_state": "idle",
                        "position": 1,
                        "notebook_uuid": uuid,
                        "uuid": cell_uuid,
                    }
                },
                "outputs": [
                    {
                        "name": "stdout",
                        "output_type": "stream",
                        "text": "Larry the Llama\n",
                    }
                ],
                "source": "__Banana Rama!__",
            }

            assert cell == expected

    @pytest.mark.asyncio
    async def test_notebook_cell_delete(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        assert cells[3]["metadata"]["jupyter_d1"]["position"] == 3

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.delete(
                f"/notebooks/{uuid}/cells/{cell_uuid}",
                headers=superuser_token_headers,
            )
            assert response.status_code == 204

            data = await receive_json(websocket)

            del cells[2]
            cells[2]["metadata"]["jupyter_d1"]["position"] = 2
            expected = {"cell_delete": cells}

            assert data == expected

    @pytest.mark.asyncio
    async def test_notebook_cell_move(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            # get the existing cells
            response = await client.get(
                f"/notebooks/{uuid}/cells", headers=superuser_token_headers
            )
            assert response.status_code == 200
            cells = response.json()["cells"]
            orig_uuids = list(
                map(lambda x: x["metadata"]["jupyter_d1"]["uuid"], cells)
            )

            # move the last cell to before the first cell
            before_uuid = orig_uuids[1]
            move_uuid = orig_uuids[3]

            params = {"before": before_uuid}
            response = await client.get(
                f"/notebooks/{uuid}/cells/{move_uuid}/move",
                query_string=params,
                headers=superuser_token_headers,
            )
            assert response.status_code == 204

            data = await receive_json(websocket)
            cells = data["cell_update"]
            assert len(cells) == 4  # should dump all the cells

            new_uuids = []
            for cell in cells:
                uuid = cell["metadata"]["jupyter_d1"]["uuid"]
                new_uuids.append(uuid)

            expected_uuids = [
                orig_uuids[0],
                orig_uuids[3],
                orig_uuids[1],
                orig_uuids[2],
            ]
            assert new_uuids == expected_uuids

            # Check cell positions
            assert_cell_positions(cells)

    @pytest.mark.asyncio
    async def test_notebook_cells_merge(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            # get the existing cells
            response = await client.get(
                f"/notebooks/{uuid}/cells", headers=superuser_token_headers
            )
            assert response.status_code == 200
            cells = response.json()["cells"]
            cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]
            cells[2]["metadata"]["jupyter_d1"]["uuid"]
            next_cell_uuid = cells[3]["metadata"]["jupyter_d1"]["uuid"]
            assert cells[1]["source"] == 'print("Larry the Llama")'
            assert cells[2]["source"] == "2+5"

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/merge",
                query_string={"above": False},
                headers=superuser_token_headers,
            )
            assert response.status_code == 204

            data = await receive_json(websocket)
            cells = data["cell_delete"]
            assert len(cells) == 3
            assert cells[1]["source"] == 'print("Larry the Llama")\n2+5'
            assert cells[1]["metadata"]["jupyter_d1"]["uuid"] == cell_uuid
            assert cells[2]["source"] == ""
            assert cells[2]["metadata"]["jupyter_d1"]["uuid"] == next_cell_uuid
            # Check cell positions
            assert_cell_positions(cells)

    @pytest.mark.asyncio
    async def test_notebook_cell_split(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            # get the existing cells
            response = await client.get(
                f"/notebooks/{uuid}/cells", headers=superuser_token_headers
            )
            assert response.status_code == 200
            cells = response.json()["cells"]
            cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]
            next_cell_uuid = cells[2]["metadata"]["jupyter_d1"]["uuid"]
            assert cells[1]["source"] == 'print("Larry the Llama")'

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/split",
                query_string={"split_location": 2},
                headers=superuser_token_headers,
            )
            assert response.status_code == 204

            data = await receive_json(websocket)
            cells = data["cell_update"]
            assert len(cells) == 5
            assert cells[1]["source"] == "pr"
            assert cells[1]["metadata"]["jupyter_d1"]["uuid"] == cell_uuid
            assert cells[2]["source"] == 'int("Larry the Llama")'
            assert cells[2]["metadata"]["jupyter_d1"]["uuid"] is not None
            assert cells[3]["metadata"]["jupyter_d1"]["uuid"] == next_cell_uuid

            # Check cell positions
            assert_cell_positions(cells)

    @pytest.mark.asyncio
    async def test_cell_stream(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]
        assert cells[1]["execution_count"] is None

        params = {
            "source": "import time\nfor i in ('heres', 'a', "
            "'streamofdata'):\n    print(i)\n    time.sleep(0.9)"
        }
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            outputs = ["heres", "a", "streamofdata"]
            cell_streams: List[Dict[str, Any]] = []
            received_info = {
                "received_cell_stream_count": 0,
                "received_cell_update_count": 0,
                "received_cell_execution": False,
                "received_vars": False,
            }

            # Receive cell stream output
            async def receive_cell_stream(data):
                cell_stream = data["cell_stream"]
                expected_output = {
                    "name": "stdout",
                    "text": f"{outputs.pop(0)}\n",
                    "output_type": "stream",
                }
                assert cell_stream["output"] == expected_output
                assert cell_stream["cell_id"] == cell_uuid
                received_info["received_cell_stream_count"] += 1
                return expected_output

            async def receive_cell_update(data):
                cell_update = data["cell_update"]
                assert len(cell_update) == 1
                cell_d1_data = cell_update[0]["metadata"]["jupyter_d1"]
                assert cell_d1_data["uuid"] == cell_uuid
                if received_info["received_cell_update_count"] == 0:
                    assert cell_d1_data["execution_state"] == "busy"
                else:
                    assert cell_d1_data["execution_state"] == "idle"
                    assert cell_update[0]["outputs"] == cell_streams
                assert (
                    cell_update[0]["execution_count"] is None
                    or cell_update[0]["execution_count"] == 1
                )
                received_info["received_cell_update_count"] += 1

            async def receive_cell_execution(data):
                cell_exec = data["cell_execution_reply"]
                assert cell_exec["cell_id"] == cell_uuid
                assert cell_exec["execution_count"] == 1
                assert len(cell_exec["parent_id"]) == 35
                received_info["received_cell_execution"] = True

            msg_types = [
                "cell_stream",
                "cell_update",
                "cell_execution_reply",
                "vars",
            ]
            for i in range(7):
                data = await receive_json(websocket, msg_types=msg_types)
                if "cell_stream" in data:
                    cell_streams.append(await receive_cell_stream(data))
                if "cell_update" in data:
                    await receive_cell_update(data)
                if "cell_execution_reply" in data:
                    await receive_cell_execution(data)
                if "vars" in data:
                    received_info["received_vars"] = True

            assert received_info["received_cell_update_count"] == 2
            assert received_info["received_cell_execution"]
            assert received_info["received_vars"]
            assert received_info["received_cell_stream_count"] == 3

    @pytest.mark.asyncio
    async def test_idle_status(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        async def is_idle_status():
            response = await client.get(
                f"/kernels/idle", headers=superuser_token_headers
            )
            assert response.status_code == 200
            json = response.json()
            return (
                json["idle"],
                datetime.strptime(json["last_idle"], "%Y-%m-%dT%H:%M:%S.%f"),
            )

        # Kernel is idle
        is_idle, last_idle_1 = await is_idle_status()
        assert is_idle is True

        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # Kernel is still idle
        is_idle, last_idle_2 = await is_idle_status()
        assert is_idle is True
        assert last_idle_1 == last_idle_2

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        busy_cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]
        quick_cell_uuid = cells[2]["metadata"]["jupyter_d1"]["uuid"]

        params = {
            "source": "from time import sleep\nwhile True:\n    sleep(5)"
        }
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{busy_cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{quick_cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            data = await receive_json(websocket, msg_types=["vars"])
            assert "vars" in data

            # Need to wait for the UPDATE_WORKDIR_REQUEST to finish
            # (and we won't see it through the websocket)
            await asyncio.sleep(1)

            # Kernel is still idle (no code running)
            is_idle, last_idle_3 = await is_idle_status()
            assert is_idle is True
            assert last_idle_3 > last_idle_2

            response = await client.get(
                f"/notebooks/{uuid}/cells/{busy_cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            # first chunk received should be a cell_update with
            # an execution_state of 'busy' when processing starts
            data = await wait_for_event(websocket, "cell_update")
            d1_metadata = data["cell_update"][0]["metadata"]["jupyter_d1"]
            assert d1_metadata["execution_state"] == "busy"

            # Kernel should not be idle now
            is_idle, last_idle_4 = await is_idle_status()
            assert is_idle is False
            assert last_idle_4 == last_idle_3

    @pytest.mark.asyncio
    async def test_kernel_interrupt(
        self, client: TestClient, superuser_token_headers: Dict[str, str],
    ):
        async def is_idle_status():
            response = await client.get(
                f"/kernels/idle", headers=superuser_token_headers
            )
            assert response.status_code == 200
            json = response.json()
            return (
                json["idle"],
                datetime.strptime(json["last_idle"], "%Y-%m-%dT%H:%M:%S.%f"),
            )

        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # Kernel is idle
        is_idle, last_idle_1 = await is_idle_status()
        assert is_idle is True

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        busy_cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]

        params = {
            "source": "from time import sleep\nwhile True:\n    sleep(5)"
        }
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{busy_cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{busy_cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            msg_id = response.json()["kernel_message"]["message_id"]

            # first chunk received should be a cell_update with
            # an execution_state of 'busy' when processing starts
            data = await wait_for_event(websocket, "cell_update")
            d1_metadata = data["cell_update"][0]["metadata"]["jupyter_d1"]
            assert d1_metadata["execution_state"] == "busy"
            assert d1_metadata["uuid"] == busy_cell_uuid

            # Kernel should not be idle now
            is_idle, last_idle_2 = await is_idle_status()
            assert is_idle is False
            assert last_idle_2 == last_idle_1

            response = await client.get(
                f"/notebooks/{uuid}/interrupt_kernel",
                headers=superuser_token_headers,
            )
            assert response.status_code == 204

            data = await wait_for_event(websocket, "kernel_interrupted")
            assert data["kernel_interrupted"]["interrupted"] is True

            # Receive cell error output
            data = await wait_for_event(websocket, "cell_update")
            d1_metadata = data["cell_update"][0]["metadata"]["jupyter_d1"]
            assert d1_metadata["execution_state"] == "busy"
            assert data["cell_update"][0]["execution_count"] is None
            assert d1_metadata["position"] == 1
            assert d1_metadata["uuid"] == busy_cell_uuid
            cell = data["cell_update"][0]
            assert cell["cell_type"] == "code"
            assert len(cell["outputs"]) == 1
            assert cell["outputs"][0]["ename"] == "KeyboardInterrupt"
            assert "KeyboardInterrupt" in cell["outputs"][0]["traceback"][1]

            def assert_json_content(data, execution_count=None):
                if "cell_update" in data:
                    cell_data = data["cell_update"][0]
                    d1_metadata = cell_data["metadata"]["jupyter_d1"]
                    assert d1_metadata["execution_state"] == "idle"
                    assert d1_metadata["uuid"] == busy_cell_uuid
                    assert cell_data["execution_count"] == execution_count
                    return None
                elif "cell_execution_reply" in data:
                    assert data["cell_execution_reply"]["execution_count"] == 1
                    assert (
                        data["cell_execution_reply"]["cell_id"]
                        == busy_cell_uuid
                    )
                    assert len(data["cell_execution_reply"]["parent_id"]) == 35
                    assert data["cell_execution_reply"]["parent_id"] == msg_id
                    return 1
                else:
                    raise AssertionError(f"unexpected cell type: {data}")

            # Execution count update
            data = await wait_for_event(websocket, "cell_execution_reply")
            exec_count = assert_json_content(data)
            data = await wait_for_event(websocket, "cell_update")
            assert_json_content(data, exec_count)

            # Kernel should be idle now
            is_idle, last_idle_3 = await is_idle_status()
            assert is_idle is True
            assert last_idle_3 > last_idle_2

    @pytest.mark.asyncio
    async def test_bogus_code_error(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]

        params = {"source": "this isnt code"}
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            msg_id = response.json()["kernel_message"]["message_id"]

            # execution_count update and busy status update,
            # in uncertain order
            def assert_json_content(data, execution_count=None):
                if "cell_update" in data:
                    cell_data = data["cell_update"][0]
                    d1_metadata = cell_data["metadata"]["jupyter_d1"]
                    assert d1_metadata["execution_state"] == "busy"
                    assert d1_metadata["uuid"] == cell_uuid
                    assert d1_metadata["position"] == 1
                    assert cell_data["execution_count"] == execution_count
                    cell = data["cell_update"][0]
                    assert cell["cell_type"] == "code"
                    assert len(cell["outputs"]) == 0
                    return None
                elif "cell_execution_reply" in data:
                    assert data["cell_execution_reply"]["execution_count"] == 1
                    assert data["cell_execution_reply"]["cell_id"] == cell_uuid
                    assert len(data["cell_execution_reply"]["parent_id"]) == 35
                    assert data["cell_execution_reply"]["parent_id"] == msg_id
                    return 1
                else:
                    raise AssertionError(f"unexpected cell type: {data}")

            msg_types = [
                "cell_update",
                "cell_execution_reply",
            ]

            data = await receive_json(websocket, msg_types=msg_types)
            exec_count = assert_json_content(data)
            data = await receive_json(websocket, msg_types=msg_types)
            assert_json_content(data, exec_count)

            # Still busy with output, error output received
            data = await receive_json(websocket, msg_types=["cell_update"])
            d1_metadata = data["cell_update"][0]["metadata"]["jupyter_d1"]
            assert d1_metadata["execution_state"] == "busy"
            assert data["cell_update"][0]["execution_count"] == 1
            assert d1_metadata["position"] == 1
            assert d1_metadata["uuid"] == cell_uuid
            cell = data["cell_update"][0]
            assert cell["cell_type"] == "code"
            assert len(cell["outputs"]) == 1
            assert cell["outputs"][0]["ename"] == "SyntaxError"
            assert "SyntaxError" in cell["outputs"][0]["traceback"][0]

            # Kernel vars update
            data = await receive_json(websocket, msg_types=["vars"])
            assert "vars" in data

    @pytest.mark.asyncio
    async def test_scratch_execute(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={"code": "4*5"},
            )
            assert response.status_code == 200
            message_id = response.json()["kernel_message"]["message_id"]

            # first chunk received should be a scratch_update with
            # an execution_state of 'busy' when processing starts
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "busy"

            # Receive scratch execution output
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            expected_output = {
                "data": {"text/plain": "20"},
                "execution_count": 1,
                "metadata": {},
                "output_type": "execute_result",
            }
            assert scratch_update["output"] == expected_output

            # Scratch execution state set to idle
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "idle"

    @pytest.mark.asyncio
    async def test_scratch_code_idle_status(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        async def is_idle_status():
            response = await client.get(
                f"/kernels/idle", headers=superuser_token_headers
            )
            assert response.status_code == 200
            json = response.json()
            return (
                json["idle"],
                datetime.strptime(json["last_idle"], "%Y-%m-%dT%H:%M:%S.%f"),
            )

        # Kernel is idle
        is_idle, last_idle_1 = await is_idle_status()
        assert is_idle is True

        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # Kernel is still idle
        is_idle, last_idle_2 = await is_idle_status()
        assert is_idle is True
        assert last_idle_1 == last_idle_2

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={"code": "4+9"},
            )
            assert response.status_code == 200
            message_id = response.json()["kernel_message"]["message_id"]

            # Execution state busy
            await receive_json(websocket, msg_types=["scratch_update"])
            # Execution output
            await receive_json(websocket, msg_types=["scratch_update"])
            # Execution state idle
            await receive_json(websocket, msg_types=["scratch_update"])

            # Kernel is still idle (no code running)
            is_idle, last_idle_3 = await is_idle_status()
            assert is_idle is True
            assert last_idle_3 > last_idle_2

            # Kernel vars update
            data = await receive_json(websocket, msg_types=["vars"])
            assert "vars" in data

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={
                    "code": "from time import sleep\n"
                    "while True:\n    sleep(1)"
                },
            )
            assert response.status_code == 200
            message_id = response.json()["kernel_message"]["message_id"]

            # first chunk received should be a cell_update with
            # an execution_state of 'busy' when processing starts
            data = await wait_for_event(websocket, "scratch_update")
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "busy"

            # Kernel should not be idle now
            is_idle, last_idle_4 = await is_idle_status()
            assert is_idle is False
            assert last_idle_4 > last_idle_3

    @pytest.mark.asyncio
    async def test_scratch_stream(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={
                    "code": "import time\nfor i in ('heres', 'a', "
                    "'streamofdata'):\n    print(i)\n    "
                    "time.sleep(0.3)"
                },
            )
            assert response.status_code == 200
            message_id = response.json()["kernel_message"]["message_id"]

            # first chunk received should be a scratch_update with
            # an execution_state of 'busy' when processing starts
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "busy"

            # Receive scratch execution output
            async def receive_text(text):
                data = await receive_json(
                    websocket, msg_types=["scratch_update"]
                )
                scratch_update = data["scratch_update"]
                expected_output = {
                    "name": "stdout",
                    "text": f"{text}\n",
                    "output_type": "stream",
                }
                assert scratch_update["output"] == expected_output

            await receive_text("heres")
            await receive_text("a")
            await receive_text("streamofdata")

            # Scratch execution state set to idle
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "idle"

    @pytest.mark.asyncio
    async def test_vars_update_cell_execute(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]

        params = {"source": "x = 7-9"}
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        assert len(response.json()["vars"]) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            # Kernel vars update
            data = await receive_json(websocket, msg_types=["vars"])
            vars = {
                var["name"]: {
                    "type": var["type"],
                    "value": var["value"],
                    "summary": var["summary"],
                }
                for var in data["vars"]
            }

            assert vars["x"]["type"] == "int"
            assert vars["x"]["value"]["single_value"] == "-2"
            assert vars["x"]["summary"] == "-2"

            response = await client.get(
                f"/notebooks/{uuid}/vars", headers=superuser_token_headers
            )
            assert response.status_code == 200
            current_vars = response.json()["vars"]
            current_vars = filter(
                lambda x: x["name"] != "___d1os", current_vars
            )
            assert list(current_vars) == data["vars"]

    @pytest.mark.asyncio
    async def test_vars_update_scratch_execute(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        assert len(response.json()["vars"]) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={"code": "x_u = 27/10"},
            )
            assert response.status_code == 200

            # Kernel vars update
            data = await receive_json(websocket, msg_types=["vars"])
            vars = {
                var["name"]: {
                    "type": var["type"],
                    "value": var["value"],
                    "summary": var["summary"],
                }
                for var in data["vars"]
            }

            assert vars["x_u"]["type"] == "float"
            assert vars["x_u"]["value"]["single_value"] == "2.7"
            assert vars["x_u"]["summary"] == "2.7"

            response = await client.get(
                f"/notebooks/{uuid}/vars", headers=superuser_token_headers
            )
            assert response.status_code == 200
            current_vars = response.json()["vars"]
            current_vars = filter(
                lambda x: x["name"] != "___d1os", current_vars
            )
            assert list(current_vars) == data["vars"]

    @pytest.mark.asyncio
    async def test_notebook_write_permission(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        readonly_token_headers: Dict[str, str],
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        with pytest.raises(AssertionError):
            async with client.websocket_connect(
                f"/notebooks/{uuid}/ws/notebook",
                headers=readonly_token_headers,
            ):
                pass

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ):
            pass

    @pytest.mark.asyncio
    async def test_notebook_stream_permission(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        with pytest.raises(AssertionError):
            async with client.websocket_connect(
                f"/notebooks/{uuid}/ws/stream",
                headers=permissionless_token_headers,
            ):
                pass

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/stream", headers=readonly_token_headers
        ):
            pass

    @pytest.mark.asyncio
    async def test_vars_update_bash(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        response = await client.post(
            "/notebooks/create",
            query_string={"kspec_name": "bash", "filename": "heyp"},
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        vars_list = response.json()["vars"]
        assert len(vars_list) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={"code": "export D1_VAR_TEST=p903s"},
            )
            assert response.status_code == 200

            data = await receive_json(
                websocket, timeout=60, msg_types=["vars"]
            )
            vars = {
                var["name"]: {
                    "type": var["type"],
                    "value": var["value"],
                    "summary": var["summary"],
                }
                for var in data["vars"]
            }
            assert vars["D1_VAR_TEST"]["type"] is None
            assert vars["D1_VAR_TEST"]["value"]["single_value"] == "p903s"
            assert vars["D1_VAR_TEST"]["summary"] == "p903s"

    @pytest.mark.asyncio
    async def test_vars_update_r_kernel(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        response = await client.post(
            "/notebooks/create",
            query_string={"kspec_name": "ir", "filename": "heyp"},
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        vars_list = response.json()["vars"]
        assert len(vars_list) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={
                    "code": "ffo <- 9\n"
                    'k = data.frame(col1=c("t", 5), '
                    'col2=c("4", 6), col3=c("h", 8))'
                },
            )
            assert response.status_code == 200

            # Give it time to install callistor
            await asyncio.sleep(10)

            data = await receive_json(
                websocket, timeout=60, msg_types=["vars"]
            )
            vars = {var["name"]: var for var in data["vars"]}

            assert vars["ffo"]["type"] == "numeric"
            assert vars["ffo"]["abbreviated"] is False
            assert vars["ffo"]["summary"] == "9"
            assert vars["ffo"]["value"]["single_value"] == "9"
            assert vars["k"]["type"] == "data.frame"
            assert vars["k"]["abbreviated"] is False
            assert "Size: 2x3 Memory: " in vars["k"]["summary"]
            assert vars["k"]["value"]["multi_value"]["column_count"] == 3
            assert vars["k"]["value"]["multi_value"]["row_count"] == 2
            assert vars["k"]["value"]["multi_value"]["column_names"] == [
                "col1 (factor)",
                "col2 (factor)",
                "col3 (factor)",
            ]
            assert vars["k"]["value"]["multi_value"]["row_names"] == ["1", "2"]
            assert vars["k"]["value"]["multi_value"]["data"] == [
                ["t", "4", "h"],
                ["5", "6", "8"],
            ]

    @pytest.mark.asyncio
    async def test_page_payload(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        response = await client.post(
            "/notebooks/create",
            query_string={"kspec_name": "python", "filename": "heyp"},
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        response = await client.post(
            f"/notebooks/{uuid}/cells",
            headers=superuser_token_headers,
            json={"cell_type": "code", "source": "?str"},
        )
        cell_uuid = response.json()["cell"]["metadata"]["jupyter_d1"]["uuid"]

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            msg_id = response.json()["kernel_message"]["message_id"]

            data = await receive_json(websocket, msg_types=["payload_page"])
            assert "text/plain" in data["payload_page"]["data"]
            assert data["payload_page"]["message_id"] == msg_id
            assert data["payload_page"]["start"] == 0

    @pytest.mark.asyncio
    async def test_code_completion(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        response = await client.post(
            "/notebooks/create",
            query_string={"kspec_name": "python", "filename": "heyp"},
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/complete",
                json={"code": "s"},
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            msg_id = response.json()["kernel_message"]["message_id"]

            data = await receive_json(websocket)
            assert len(data["complete_reply"]["matches"]) > 0
            assert data["complete_reply"]["message_id"] == msg_id
            assert data["complete_reply"]["cursor_start"] == 0
            assert data["complete_reply"]["cursor_end"] == 1

    @pytest.mark.asyncio
    async def test_code_completion_cursor_position(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        response = await client.post(
            "/notebooks/create",
            query_string={"kspec_name": "python", "filename": "heyp"},
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/complete",
                json={"code": "5 + prin + 6", "cursor_position": 7},
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            msg_id = response.json()["kernel_message"]["message_id"]

            data = await receive_json(websocket)
            assert data["complete_reply"]["matches"] == ["print"]
            assert data["complete_reply"]["message_id"] == msg_id
            assert data["complete_reply"]["cursor_start"] == 4
            assert data["complete_reply"]["cursor_end"] == 7

    @pytest.mark.asyncio
    async def test_history(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            await receive_json(websocket)
            await receive_json(websocket)
            await receive_json(websocket)
            await receive_json(websocket)
            await receive_json(websocket)

            response = await client.post(
                f"/notebooks/{uuid}/history", headers=superuser_token_headers
            )
            assert response.status_code == 200
            msg_id = response.json()["kernel_message"]["message_id"]

            data = await wait_for_event(websocket, "history_reply")

            assert data["history_reply"]["message_id"] == msg_id
            history = data["history_reply"]["history"]
            assert history[0]["line_number"] == 0
            assert history[0]["input"] == ""
            assert history[0]["output"] is None
            assert history[1]["line_number"] == 1
            assert history[1]["input"] == "2+5"
            assert history[1]["output"] == "7"

    @pytest.mark.asyncio
    async def test_restart_kernel(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        response = await client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        assert cells[2]["outputs"][0]["data"]["text/plain"] == "7"

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:

            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/restart_kernel",
                headers=superuser_token_headers,
            )

            assert response.status_code == 204

            data = await wait_for_event(websocket, "kernel_restarted")

            assert data["kernel_restarted"]["run_all_cells"] is False
            cells = data["kernel_restarted"]["cells"]
            assert cells[2]["outputs"][0]["data"]["text/plain"] == "7"
            assert cells[2]["execution_count"] == 1

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        for cell in cells:
            assert cells[2]["outputs"][0]["data"]["text/plain"] == "7"
            assert cells[2]["execution_count"] == 1

    @pytest.mark.asyncio
    async def test_restart_kernel_and_clear_output(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        response = await client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        assert cells[2]["outputs"][0]["data"]["text/plain"] == "7"

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:

            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/restart_kernel_and_clear_output",
                headers=superuser_token_headers,
            )

            assert response.status_code == 204

            data = await wait_for_event(websocket, "kernel_restarted")

            assert data["kernel_restarted"]["run_all_cells"] is False
            cells = data["kernel_restarted"]["cells"]
            for cell in cells:
                assert (
                    ("outputs" not in cell or len(cell["outputs"]) == 0)
                    and "execution_count" not in cell
                    or cell["execution_count"] is None
                )

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        for cell in cells:
            assert (
                ("outputs" not in cell or len(cell["outputs"]) == 0)
                and "execution_count" not in cell
                or cell["execution_count"] is None
            )

    @pytest.mark.asyncio
    async def test_restart_kernel_and_run_all_cells(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(client, superuser_token_headers)

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        response = await client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        assert "outputs" not in cells[0]
        assert cells[1]["outputs"][0]["text"] == "Larry the Llama\n"
        assert cells[2]["outputs"][0]["data"]["text/plain"] == "7"

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:

            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            # Clear the web socket before we start
            await collect_websocket_messages(websocket, timeout=2)

            response = await client.get(
                f"/notebooks/{uuid}/restart_kernel_and_execute_all",
                headers=superuser_token_headers,
            )

            assert response.status_code == 200

            msgs = await collect_websocket_messages(websocket, timeout=10)
            kernel_restarted_msgs = filter_websocket_collection(
                msgs, "kernel_restarted"
            )
            cell_update_msgs = filter_websocket_collection(msgs, "cell_update")

            data = kernel_restarted_msgs.pop(0)
            assert data["kernel_restarted"]["run_all_cells"] is True
            cells = data["kernel_restarted"]["cells"]
            for cell in cells:
                assert (
                    ("outputs" not in cell or len(cell["outputs"]) == 0)
                    and "execution_count" not in cell
                    or cell["execution_count"] is None
                )

            data = cell_update_msgs.pop(0)
            cell = data["cell_update"][0]
            assert len(cell["outputs"]) == 0
            assert cell["metadata"]["jupyter_d1"]["execution_state"] == "busy"

            data = cell_update_msgs.pop(0)
            cell = data["cell_update"][0]
            assert cell["outputs"][0]["text"] == "Larry the Llama\n"
            assert cell["execution_count"] == 1
            assert cell["metadata"]["jupyter_d1"]["execution_state"] == "idle"

            data = cell_update_msgs.pop(0)
            cell = data["cell_update"][0]
            assert len(cell["outputs"]) == 0
            assert cell["metadata"]["jupyter_d1"]["execution_state"] == "busy"

            data = cell_update_msgs.pop(0)
            cell = data["cell_update"][0]
            assert cell["outputs"][0]["data"]["text/plain"] == "7"
            assert cell["execution_count"] == 2
            assert cell["metadata"]["jupyter_d1"]["execution_state"] == "busy"

            data = cell_update_msgs.pop(0)
            cell = data["cell_update"][0]
            assert cell["outputs"][0]["data"]["text/plain"] == "7"
            assert cell["execution_count"] == 2
            assert cell["metadata"]["jupyter_d1"]["execution_state"] == "idle"

    @pytest.mark.asyncio
    async def test_ipython_kernel_startup_script(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Test that the callisto_startup.py startup script is run when the
        kernel starts.
        """
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]

        params = {
            "source": "import pandas as pd\n"
            'pd.get_option("display.html.table_schema")'
        }
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        assert len(response.json()["vars"]) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.patch(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                json=params,
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            expected = {
                "cell_update": [
                    {
                        "cell_type": "code",
                        "execution_count": 1,
                        "metadata": {
                            "jupyter_d1": {
                                "position": 1,
                                "uuid": cell_uuid,
                                "notebook_uuid": uuid,
                                "execution_state": "busy",
                            }
                        },
                        "outputs": [
                            {
                                "data": {"text/plain": "True"},
                                "execution_count": 1,
                                "metadata": {},
                                "output_type": "execute_result",
                            }
                        ],
                        "source": "import pandas as pd\n"
                        'pd.get_option("display.html.table_schema")',
                    }
                ]
            }

            patch_json = response.json()
            # response should contain both the 'cell','kernel_message' wrappers
            assert "cell" in patch_json.keys()
            msg_id = response.json()["kernel_message"]["message_id"]
            assert len(msg_id) == 35

            exec_reply = await receive_json(
                websocket, msg_types=["cell_execution_reply"]
            )
            assert exec_reply["cell_execution_reply"]["cell_id"] == cell_uuid
            assert exec_reply["cell_execution_reply"]["parent_id"] == msg_id

            data = await receive_json(websocket, msg_types=["cell_update"])
            assert data == expected


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestKernelVarsManager:
    @pytest.mark.asyncio
    async def test_get_single_var_detail(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Large dataframe is abbreviated in summary, not in variable
        details request.
        """
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]

        params = {
            "source": "import pandas as pd\ndf3 = "
            "pd.DataFrame([{str(i): i + j for i in range(5)} for j in "
            "range(150)])"
        }
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        assert len(response.json()["vars"]) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            # Kernel vars update
            data = await receive_json(websocket, msg_types=["vars"])
            vars = {
                var["name"]: {
                    "type": var["type"],
                    "value": var["value"],
                    "summary": var["summary"],
                    "abbreviated": var["abbreviated"],
                }
                for var in data["vars"]
            }

            assert vars["df3"]["type"] == "DataFrame"
            assert vars["df3"]["value"]["multi_value"]["column_count"] == 5
            assert vars["df3"]["value"]["multi_value"]["row_count"] == 150
            assert vars["df3"]["value"]["multi_value"]["column_names"] == [
                str(i) for i in range(5)
            ]
            assert vars["df3"]["value"]["multi_value"]["data"] == [
                [str(i + j) for i in range(5)] for j in range(10)
            ]
            assert vars["df3"]["summary"] == "Size: 150x5 Memory: 5.98 KB"
            assert vars["df3"]["abbreviated"] is True

            response = await client.get(
                f"/notebooks/{uuid}/vars", headers=superuser_token_headers
            )
            assert response.status_code == 200
            current_vars = response.json()["vars"]
            current_vars = filter(
                lambda x: x["name"] != "___d1os", current_vars
            )
            assert list(current_vars) == data["vars"]

            response = await client.get(
                f"/notebooks/{uuid}/vars/df3", headers=superuser_token_headers
            )
            assert response.status_code == 200
            single_var = response.json()["single_var"]
            assert single_var["name"] == "df3"
            assert single_var["type"] == "DataFrame"
            assert single_var["value"]["multi_value"]["column_count"] == 5
            assert single_var["value"]["multi_value"]["row_count"] == 150
            assert single_var["value"]["multi_value"]["column_names"] == [
                str(i) for i in range(5)
            ]
            assert single_var["value"]["multi_value"]["data"] == [
                [str(i + j) for i in range(5)] for j in range(150)
            ]
            assert single_var["summary"] == "Size: 150x5 Memory: 5.98 KB"
            assert single_var["abbreviated"] is False

            # Shouldn't have changed anything in the current vars
            response = await client.get(
                f"/notebooks/{uuid}/vars", headers=superuser_token_headers
            )
            assert response.status_code == 200
            current_vars = response.json()["vars"]
            current_vars = filter(
                lambda x: x["name"] != "___d1os", current_vars
            )
            assert list(current_vars) == data["vars"]

            # Var that doesn't exist
            response = await client.get(
                f"/notebooks/{uuid}/vars/df2", headers=superuser_token_headers
            )
            assert response.status_code == 404

    @pytest.mark.asyncio
    async def test_get_single_var_detail_r_kernel(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        R Kernel large dataframe is abbreviated in summary, not in variable
        details request.
        """
        uuid = await upload_notebook(
            client, superuser_token_headers, "simple_r.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == "f <- 543"

        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]

        params = {"source": "kool = data.frame(col1=c(1:150), col2=c(150:1))"}
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}/vars", headers=superuser_token_headers
        )
        assert response.status_code == 200
        assert len(response.json()["vars"]) == 0

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            # Kernel vars update
            data = await receive_json(websocket, msg_types=["vars"])
            vars = {var["name"]: var for var in data["vars"]}

            assert vars["kool"]["type"] == "data.frame"
            assert vars["kool"]["value"]["multi_value"]["column_count"] == 2
            assert vars["kool"]["value"]["multi_value"]["row_count"] == 150
            assert vars["kool"]["value"]["multi_value"]["column_names"] == [
                "col1 (integer)",
                "col2 (integer)",
            ]
            assert vars["kool"]["value"]["multi_value"]["data"] == [
                [str(i + 1), str(150 - i)] for i in range(25)
            ]
            assert "Size: 150x2 Memory:" in vars["kool"]["summary"]
            assert vars["kool"]["abbreviated"] is True

            response = await client.get(
                f"/notebooks/{uuid}/vars", headers=superuser_token_headers
            )
            assert response.status_code == 200
            current_vars = response.json()["vars"]
            assert list(current_vars) == data["vars"]

            response = await client.get(
                f"/notebooks/{uuid}/vars/kool", headers=superuser_token_headers
            )
            assert response.status_code == 200
            single_var = response.json()["single_var"]
            assert single_var["name"] == "kool"
            assert single_var["type"] == "data.frame"
            assert single_var["value"]["multi_value"]["column_count"] == 2
            assert single_var["value"]["multi_value"]["row_count"] == 150
            assert single_var["value"]["multi_value"]["column_names"] == [
                "col1 (integer)",
                "col2 (integer)",
            ]
            assert single_var["value"]["multi_value"]["data"] == [
                [str(i + 1), str(150 - i)] for i in range(150)
            ]
            assert "Size: 150x2 Memory:" in single_var["summary"]
            assert single_var["abbreviated"] is False

            # Shouldn't have changed anything in the current vars
            response = await client.get(
                f"/notebooks/{uuid}/vars", headers=superuser_token_headers
            )
            assert response.status_code == 200
            current_vars = response.json()["vars"]
            assert list(current_vars) == data["vars"]

            # Var that doesn't exist
            response = await client.get(
                f"/notebooks/{uuid}/vars/df2", headers=superuser_token_headers
            )
            assert response.status_code == 404


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestWorkingDirectory:
    async def assert_python_working_directory(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        open_path: str,
        working_directory: str,
        query_string: Optional[Dict[str, Any]] = None,
    ):
        """Test that python kernel is started in correct directory."""
        if query_string is None:
            query_string = {}
        query_string["filepath"] = open_path
        response = await client.get(
            f"/notebooks/open/",
            headers=superuser_token_headers,
            query_string=query_string,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        # /tmp symlinks to /private/tmp on mac, so we resolve
        resolved_work_dir = str(pathlib.Path(working_directory).resolve())

        assert (
            resp_json["metadata"]["jupyter_d1"]["working_directory"]
            == resolved_work_dir
        )

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={"code": "import os\nos.getcwd()"},
            )
            assert response.status_code == 200
            message_id = response.json()["kernel_message"]["message_id"]

            # first chunk received should be a scratch_update with
            # an execution_state of 'busy' when processing starts
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "busy"

            # Receive scratch execution output
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]

            expected_output = {
                "data": {"text/plain": f"'{resolved_work_dir}'"},
                "execution_count": 1,
                "metadata": {},
                "output_type": "execute_result",
            }
            assert scratch_update["output"] == expected_output

            # Scratch execution state set to idle
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "idle"

            token = get_superuser_token()
            auth = base64.b64encode(f"{token}:".encode("utf-8"))
            response = await client.get(
                f"/dav{resolved_work_dir}",
                headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
            )
            assert response.status_code == 200

    @pytest.mark.asyncio
    async def test_working_directory_passed_in(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        If notebook is opened with working directory specified, make sure
        the kernel is running in that directory.
        """
        filename = "simple.ipynb"
        nb_filename = f"jupyter_d1/tests/notebooks/{filename}"
        nb_json = open(nb_filename).read()
        response = await client.post(
            "/notebooks/upload",
            query_string={"filename": filename},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        await self.assert_python_working_directory(
            client,
            superuser_token_headers,
            open_path=filename,
            working_directory=f"{os.getcwd()}/jupyter_d1/tests/notebooks",
            query_string={
                "working_directory": f"{os.getcwd()}/jupyter_d1"
                "/tests/notebooks"
            },
        )

    @pytest.mark.asyncio
    async def test_file_in_different_working_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        If notebook in a different directory is opened, make sure
        the kernel is running in that directory.
        """
        shutil.copyfile(
            "jupyter_d1/tests/notebooks/simple.ipynb",
            "jupyter_d1/tests/notebooks/simple_copy.ipynb",
        )
        await self.assert_python_working_directory(
            client,
            superuser_token_headers,
            open_path=f"{os.getcwd()}/jupyter_d1/tests/"
            "notebooks/simple_copy.ipynb",
            working_directory=f"{os.getcwd()}/jupyter_d1/tests/notebooks",
        )
        os.remove("jupyter_d1/tests/notebooks/simple_copy.ipynb")

    @pytest.mark.asyncio
    async def test_file_in_parent_directory_of_working_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        If notebook in the parent directory of the server's working directory 
        is opened, make sure the kernel is running in that parent directory.
        """
        shutil.copyfile(
            "jupyter_d1/tests/notebooks/simple.ipynb",
            f"/tmp/simple_copy.ipynb",
        )
        await self.assert_python_working_directory(
            client,
            superuser_token_headers,
            open_path=f"/tmp/simple_copy.ipynb",
            working_directory=f"/tmp",
        )
        os.remove("/tmp/simple_copy.ipynb")

    @pytest.mark.asyncio
    async def test_file_in_sub_directory_of_working_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        If notebook in a sub directory of the server's working directory 
        is opened, make sure the kernel is running in that sub directory.
        """
        pathlib.Path("/tmp/jupyter_d1_test/some_nbs").mkdir(exist_ok=True)
        shutil.copyfile(
            "jupyter_d1/tests/notebooks/simple.ipynb",
            f"/tmp/jupyter_d1_test/some_nbs//simple_copy.ipynb",
        )
        await self.assert_python_working_directory(
            client,
            superuser_token_headers,
            open_path=f"/tmp/jupyter_d1_test/some_nbs/simple_copy.ipynb",
            working_directory=f"/tmp/jupyter_d1_test/some_nbs",
        )
        shutil.rmtree("/tmp/jupyter_d1_test/some_nbs")

    @pytest.mark.asyncio
    async def test_r_kernel_working_directory_passed_in(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        working_directory = "/tmp"
        response = await client.post(
            "/notebooks/create",
            query_string={
                "kspec_name": "ir",
                "filename": "heyp",
                "working_directory": working_directory,
            },
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={"code": "getwd()"},
            )
            assert response.status_code == 200
            message_id = response.json()["kernel_message"]["message_id"]

            # first chunk received should be a scratch_update with
            # an execution_state of 'busy' when processing starts
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "busy"

            # Receive scratch execution output
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            # /tmp symlinks to /private/tmp on mac, so we resolve
            resolved_work_dir = pathlib.Path(working_directory).resolve()
            expected_output = {
                "data": {
                    "text/html": f"'{resolved_work_dir}'",
                    "text/latex": f"'{resolved_work_dir}'",
                    "text/markdown": f"'{resolved_work_dir}'",
                    "text/plain": f'[1] "{resolved_work_dir}"',
                },
                "metadata": {},
                "output_type": "display_data",
            }
            assert scratch_update["output"] == expected_output

            # Scratch execution state set to idle
            data = await receive_json(websocket, msg_types=["scratch_update"])
            scratch_update = data["scratch_update"]
            assert scratch_update["message_id"] == message_id
            assert scratch_update["execution_state"] == "idle"

            token = get_superuser_token()
            auth = base64.b64encode(f"{token}:".encode("utf-8"))
            response = await client.get(
                f"/dav{resolved_work_dir}",
                headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
            )
            assert response.status_code == 200


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestChangeWorkingDirectory:
    async def do_test_change_working_directory(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        kspec_name: str,
    ):
        working_directory = "/tmp"
        resolved_work_dir = str(pathlib.Path(working_directory).resolve())
        response = await client.post(
            "/notebooks/create",
            query_string={
                "kspec_name": kspec_name,
                "filename": "heyp",
                "working_directory": working_directory,
            },
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        token = get_superuser_token()
        auth = base64.b64encode(f"{token}:".encode("utf-8"))
        response = await client.get(
            f"/dav{resolved_work_dir}",
            headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
        )
        assert response.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            new_workdir = f"{os.getcwd()}/jupyter_d1/tests/notebooks"
            response = await client.patch(
                f"/notebooks/{uuid}/change_working_directory",
                headers=superuser_token_headers,
                query_string={"directory": new_workdir},
            )
            assert response.status_code == 200

            data = await receive_json(websocket, msg_types=["metadata"])

            metadata = data["metadata"]
            assert metadata["jupyter_d1"]["working_directory"] == new_workdir

            response = await client.get(
                f"/dav{new_workdir}",
                headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
            )
            assert response.status_code == 200

            response = await client.get(
                f"/dav{resolved_work_dir}",
                headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
            )
            assert response.status_code == 404

        response = await client.get(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        resp_json = response.json()["notebook"]
        assert (
            resp_json["metadata"]["jupyter_d1"]["working_directory"]
            == new_workdir
        )

    async def do_test_change_working_directory_in_code(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        kspec_name: str,
        chdir_code: str,
    ):
        working_directory = "/tmp"
        resolved_work_dir = str(pathlib.Path(working_directory).resolve())
        response = await client.post(
            "/notebooks/create",
            query_string={
                "kspec_name": kspec_name,
                "filename": "heyp",
                "working_directory": working_directory,
            },
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        token = get_superuser_token()
        auth = base64.b64encode(f"{token}:".encode("utf-8"))
        response = await client.get(
            f"/dav{resolved_work_dir}",
            headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
        )
        assert response.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            new_workdir = f"{os.getcwd()}/jupyter_d1/tests/notebooks"
            response = await client.post(
                f"/notebooks/{uuid}/scratch_execute",
                headers=superuser_token_headers,
                json={"code": chdir_code.format(new_workdir=new_workdir)},
            )
            assert response.status_code == 200

            data = await wait_for_event(websocket, "metadata", timeout=10)

            metadata = data["metadata"]
            assert metadata["jupyter_d1"]["working_directory"] == new_workdir

            response = await client.get(
                f"/dav{new_workdir}",
                headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
            )
            assert response.status_code == 200

            response = await client.get(
                f"/dav{resolved_work_dir}",
                headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
            )
            assert response.status_code == 404

        response = await client.get(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        resp_json = response.json()["notebook"]
        assert (
            resp_json["metadata"]["jupyter_d1"]["working_directory"]
            == new_workdir
        )

    @pytest.mark.asyncio
    async def test_python_change_working_directory_endpoint(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        await self.do_test_change_working_directory(
            client, superuser_token_headers, "python"
        )

    @pytest.mark.asyncio
    async def test_python_change_working_directory_in_code(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        await self.do_test_change_working_directory_in_code(
            client,
            superuser_token_headers,
            "python",
            'import os\nos.chdir("{new_workdir}")',
        )

    @pytest.mark.asyncio
    async def test_r_change_working_directory_endpoint(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        await self.do_test_change_working_directory(
            client, superuser_token_headers, "ir"
        )

    @pytest.mark.asyncio
    async def test_r_change_working_directory_in_code(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        await self.do_test_change_working_directory_in_code(
            client, superuser_token_headers, "ir", 'setwd("{new_workdir}")'
        )

    @pytest.mark.asyncio
    async def test_bash_change_working_directory_endpoint(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        await self.do_test_change_working_directory(
            client, superuser_token_headers, "bash"
        )

    @pytest.mark.asyncio
    async def test_bash_change_working_directory_in_code(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        await self.do_test_change_working_directory_in_code(
            client, superuser_token_headers, "bash", "cd {new_workdir}"
        )

    @pytest.mark.asyncio
    async def test_change_working_directory_fail_invalid_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        working_directory = "/tmp"
        resolved_work_dir = str(pathlib.Path(working_directory).resolve())
        response = await client.post(
            "/notebooks/create",
            query_string={
                "kspec_name": "python",
                "filename": "heyp",
                "working_directory": working_directory,
            },
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()["notebook"]
        uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]

        token = get_superuser_token()
        auth = base64.b64encode(f"{token}:".encode("utf-8"))
        response = await client.get(
            f"/dav{resolved_work_dir}",
            headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
        )
        assert response.status_code == 200

        await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

        new_workdir = f"{os.getcwd()}/tests/notrealdir"
        response = await client.patch(
            f"/notebooks/{uuid}/change_working_directory",
            headers=superuser_token_headers,
            query_string={"directory": new_workdir},
        )
        assert response.status_code == 404
        assert response.json()["detail"] == "Directory does not exist"

        await asyncio.sleep(2)

        response = await client.get(
            f"/dav{new_workdir}",
            headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
        )
        assert response.status_code == 404

        response = await client.get(
            f"/dav{resolved_work_dir}",
            headers={"Authorization": f"Basic {auth.decode('utf-8')}"},
        )
        assert response.status_code == 200

        response = await client.get(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        resp_json = response.json()["notebook"]
        assert (
            resp_json["metadata"]["jupyter_d1"]["working_directory"]
            == resolved_work_dir
        )


@pytest.mark.usefixtures(
    "clear_notebooks", "clear_notebook_directory", "clear_notebook_directory_2"
)
class TestSaveAs:
    @pytest.mark.asyncio
    async def test_rename(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, filename="simple.ipynb"
        )

        assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")
        assert not os.path.exists(f"{settings.ROOT_DIR}/awholenewname.ipynb")

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            response = await client.patch(
                f"/notebooks/{uuid}/rename",
                query_string={"filename": "awholenewname"},
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            assert (
                response.json()["path"]
                == f"{settings.ROOT_DIR}/awholenewname.ipynb"
            )

            data = await receive_json(websocket, msg_types=["metadata"])
            metadata = data["metadata"]
            assert (
                metadata["jupyter_d1"]["path"]
                == f"{settings.ROOT_DIR}/awholenewname.ipynb"
            )
            assert metadata["jupyter_d1"]["name"] == "awholenewname"

            assert os.path.exists(f"{settings.ROOT_DIR}/awholenewname.ipynb")
            assert not os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")

    @pytest.mark.asyncio
    async def test_rename_different_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, filename="simple.ipynb"
        )

        assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")
        assert not os.path.exists(f"{workdir_1}/awholenewname.ipynb")

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            response = await client.patch(
                f"/notebooks/{uuid}/rename",
                query_string={
                    "filename": "awholenewname",
                    "directory": "/tmp/test_nb_2/",
                },
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            assert (
                response.json()["path"]
                == f"/tmp/test_nb_2/awholenewname.ipynb"
            )

            data = await receive_json(websocket, msg_types=["metadata"])
            metadata = data["metadata"]
            assert (
                metadata["jupyter_d1"]["path"]
                == f"{workdir_1}/awholenewname.ipynb"
            )
            assert metadata["jupyter_d1"]["name"] == "awholenewname"

            assert os.path.exists(f"{workdir_1}/awholenewname.ipynb")
            assert not os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")

    @pytest.mark.asyncio
    async def test_rename_file_exists(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """Fail if file exists unless overwrite param is set to true"""
        uuid = await upload_notebook(
            client, superuser_token_headers, filename="simple.ipynb"
        )

        with open(f"{workdir_1}/awholenewname.ipynb", "w") as f:
            f.write("some text here")

        assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")
        assert os.path.exists(f"{workdir_1}/awholenewname.ipynb")

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            response = await client.patch(
                f"/notebooks/{uuid}/rename",
                query_string={
                    "filename": "awholenewname",
                    "directory": "/tmp/test_nb_2/",
                },
                headers=superuser_token_headers,
            )
            assert response.status_code == 400
            assert response.json()["detail"] == f"File already exists"

            response = await client.patch(
                f"/notebooks/{uuid}/rename",
                query_string={
                    "filename": "awholenewname",
                    "directory": "/tmp/test_nb_2/",
                    "overwrite": True,
                },
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            assert (
                response.json()["path"]
                == f"/tmp/test_nb_2/awholenewname.ipynb"
            )

            data = await receive_json(websocket, msg_types=["metadata"])
            metadata = data["metadata"]
            assert (
                metadata["jupyter_d1"]["path"]
                == f"{workdir_1}/awholenewname.ipynb"
            )
            assert metadata["jupyter_d1"]["name"] == "awholenewname"

            assert os.path.exists(f"{workdir_1}/awholenewname.ipynb")
            assert not os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")

        with open(f"{workdir_1}/awholenewname.ipynb", "r") as f:
            assert f.read() != "some text here"

    @pytest.mark.asyncio
    async def test_save_as(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, filename="simple.ipynb"
        )

        assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")
        assert not os.path.exists(f"{settings.ROOT_DIR}/awholenewname.ipynb")

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            response = await client.patch(
                f"/notebooks/{uuid}/save_as",
                query_string={"filename": "awholenewname"},
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            assert (
                response.json()["path"]
                == f"{settings.ROOT_DIR}/awholenewname.ipynb"
            )

            data = await receive_json(websocket, msg_types=["metadata"])
            metadata = data["metadata"]
            assert (
                metadata["jupyter_d1"]["path"]
                == f"{settings.ROOT_DIR}/awholenewname.ipynb"
            )
            assert metadata["jupyter_d1"]["name"] == "awholenewname"
            assert metadata["jupyter_d1"]["uuid"] == uuid

            assert os.path.exists(f"{settings.ROOT_DIR}/awholenewname.ipynb")
            assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")

            response = await client.get(
                f"/notebooks/{uuid}", headers=superuser_token_headers
            )
            resp_json = response.json()["notebook"]
            assert (
                resp_json["metadata"]["jupyter_d1"]["name"] == "awholenewname"
            )
            assert resp_json["metadata"]["jupyter_d1"]["uuid"] == uuid

    @pytest.mark.asyncio
    async def test_save_as_different_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = await upload_notebook(
            client, superuser_token_headers, filename="simple.ipynb"
        )

        assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")
        assert not os.path.exists(f"{workdir_1}/awholenewname.ipynb")

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            response = await client.patch(
                f"/notebooks/{uuid}/save_as",
                query_string={
                    "filename": "awholenewname",
                    "directory": "/tmp/test_nb_2/",
                },
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            assert (
                response.json()["path"]
                == f"/tmp/test_nb_2/awholenewname.ipynb"
            )

            data = await receive_json(websocket, msg_types=["metadata"])
            metadata = data["metadata"]
            assert (
                metadata["jupyter_d1"]["path"]
                == f"{workdir_1}/awholenewname.ipynb"
            )
            assert metadata["jupyter_d1"]["name"] == "awholenewname"

            assert os.path.exists(f"{workdir_1}/awholenewname.ipynb")
            assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")

    @pytest.mark.asyncio
    async def test_save_as_file_exists(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """Fail if file exists unless overwrite param is set to true"""
        uuid = await upload_notebook(
            client, superuser_token_headers, filename="simple.ipynb"
        )

        with open(f"{workdir_1}/awholenewname.ipynb", "w") as f:
            f.write("some text here")

        assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")
        assert os.path.exists(f"{workdir_1}/awholenewname.ipynb")

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            response = await client.patch(
                f"/notebooks/{uuid}/save_as",
                query_string={
                    "filename": "awholenewname",
                    "directory": "/tmp/test_nb_2/",
                },
                headers=superuser_token_headers,
            )
            assert response.status_code == 400
            assert response.json()["detail"] == f"File already exists"

            response = await client.patch(
                f"/notebooks/{uuid}/save_as",
                query_string={
                    "filename": "awholenewname",
                    "directory": "/tmp/test_nb_2/",
                    "overwrite": True,
                },
                headers=superuser_token_headers,
            )
            assert response.status_code == 200
            assert (
                response.json()["path"]
                == f"/tmp/test_nb_2/awholenewname.ipynb"
            )

            data = await receive_json(websocket, msg_types=["metadata"])
            metadata = data["metadata"]
            assert (
                metadata["jupyter_d1"]["path"]
                == f"{workdir_1}/awholenewname.ipynb"
            )
            assert metadata["jupyter_d1"]["name"] == "awholenewname"

            assert os.path.exists(f"{workdir_1}/awholenewname.ipynb")
            assert os.path.exists(f"{settings.ROOT_DIR}/simple.ipynb")

        with open(f"{workdir_1}/awholenewname.ipynb", "r") as f:
            assert f.read() != "some text here"


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestD1Commands:
    async def do_test_d1_notify(
        self,
        cell_content: str,
        notifications: List[str],
        remaining_stdout: str,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        mocker: MockFixture,
    ):
        execute_d1_command = mocker.patch(
            "jupyter_d1.storage.notebook_manager.execute_d1_command"
        )

        uuid = await upload_notebook(
            client, superuser_token_headers, "simple.ipynb"
        )

        # get an existing cell
        response = await client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert cells[1]["source"] == 'print("Larry the Llama")'

        cell_uuid = cells[1]["metadata"]["jupyter_d1"]["uuid"]
        assert cells[1]["execution_count"] is None

        body = {"source": cell_content}
        response = await client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=body,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        async with client.websocket_connect(
            f"/notebooks/{uuid}/ws/notebook", headers=superuser_token_headers
        ) as websocket:
            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            response = await client.get(
                f"/notebooks/{uuid}/cells/{cell_uuid}/execute",
                headers=superuser_token_headers,
            )
            assert response.status_code == 200

            cell_streams: List[Dict[str, Any]] = []
            received_info = {
                "received_cell_stream_count": 0,
                "received_cell_update_count": 0,
                "received_cell_execution": False,
                "received_vars": False,
            }

            # Receive cell stream output
            async def receive_cell_stream(data):
                cell_stream = data["cell_stream"]
                expected_output = {
                    "name": "stdout",
                    "text": remaining_stdout,
                    "output_type": "stream",
                }
                assert cell_stream["output"] == expected_output
                assert cell_stream["cell_id"] == cell_uuid
                received_info["received_cell_stream_count"] += 1
                return expected_output

            async def receive_cell_update(data):
                cell_update = data["cell_update"]
                assert len(cell_update) == 1
                cell_d1_data = cell_update[0]["metadata"]["jupyter_d1"]
                assert cell_d1_data["uuid"] == cell_uuid
                if received_info["received_cell_update_count"] == 0:
                    assert cell_d1_data["execution_state"] == "busy"
                    assert (
                        cell_update[0]["execution_count"] is None
                        or cell_update[0]["execution_count"] == 1
                    )
                else:
                    assert cell_d1_data["execution_state"] == "idle"
                    assert cell_update[0]["outputs"] == cell_streams
                    assert cell_update[0]["execution_count"] == 1
                received_info["received_cell_update_count"] += 1

            async def receive_cell_execution(data):
                cell_exec = data["cell_execution_reply"]
                assert cell_exec["cell_id"] == cell_uuid
                assert cell_exec["execution_count"] == 1
                assert len(cell_exec["parent_id"]) == 35
                received_info["received_cell_execution"] = True

            msg_types = [
                "cell_stream",
                "cell_update",
                "cell_execution_reply",
                "vars",
            ]
            for i in range(5):
                data = await receive_json(websocket, 20, msg_types=msg_types)
                if "cell_stream" in data:
                    cell_streams.append(await receive_cell_stream(data))
                if "cell_update" in data:
                    await receive_cell_update(data)
                if "cell_execution_reply" in data:
                    await receive_cell_execution(data)
                if "vars" in data:
                    received_info["received_vars"] = True

            assert received_info["received_cell_update_count"] == 2
            assert received_info["received_cell_execution"]
            assert received_info["received_vars"]
            assert received_info["received_cell_stream_count"] == 1

            execute_d1_command.assert_has_calls(
                [call(note) for note in notifications]
            )

    @pytest.mark.asyncio
    async def test_d1_notify(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        mocker: MockFixture,
    ):
        cell_content = (
            'print("___callisto_d1_command___{'
            '\\"command_type\\":\\"notify\\",'
            '\\"title\\":\\"atitle\\",'
            '\\"message\\":\\"test6e4\\"}___callisto_d1_command___", '
            ' end="")'
        )
        notifications = [
            '{"command_type":"notify",'
            '"title":"atitle",'
            '"message":"test6e4"}'
        ]
        await self.do_test_d1_notify(
            cell_content,
            notifications,
            "",
            client,
            superuser_token_headers,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_d1_notify_incomplete(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        mocker: MockFixture,
    ):
        cell_content = (
            'print("ruh___callisto_d1_command___{'
            '\\"command_type\\":\\"notify\\",'
            '\\"title\\":\\"atitle\\",'
            '\\"message\\":\\"test6e4\\"}___callisto_d1_command___mori'
            '___callisto_d1_command___jjuo", '
            ' end="")'
        )
        notifications = [
            '{"command_type":"notify",'
            '"title":"atitle",'
            '"message":"test6e4"}'
        ]
        await self.do_test_d1_notify(
            cell_content,
            notifications,
            "ruhmorijjuo",
            client,
            superuser_token_headers,
            mocker,
        )

    @pytest.mark.asyncio
    async def test_d1_notify_invalid_command(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        mocker: MockFixture,
    ):
        cell_content = (
            'print("___callisto_d1_command___{'
            '\\"command_type\\":\\"notify\\",'
            '\\"title\\":\\"atitle\\",'
            '\\"message\\":\\"test6e4\\"}___callisto_d1_command___mori'
            "___callisto_d1_command___{jjuo}___callisto_d1_command___"
            'whey", '
            ' end="")'
        )
        notifications = [
            '{"command_type":"notify",'
            '"title":"atitle",'
            '"message":"test6e4"}',
            "{jjuo}",
        ]
        await self.do_test_d1_notify(
            cell_content,
            notifications,
            "moriwhey",
            client,
            superuser_token_headers,
            mocker,
        )


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestNotebooksWebSocket:
    async def get_notebooks_event(self, websocket):
        for i in range(5):
            data = await receive_json(websocket)
            if "notebooks" in data:
                break
        return data

    @pytest.mark.asyncio
    async def test_notebooks_updates(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):

        async with client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:

            await asyncio.sleep(WEBSOCKET_SLEEP_TIME)

            uuid = await upload_notebook(client, superuser_token_headers)

            data = await wait_for_event(websocket, "notebooks", tries=100)
            assert len(data["notebooks"]) == 1
            assert data["notebooks"][0]["uuid"] == uuid
            assert data["notebooks"][0]["name"] == "simple"

            uuid2 = await upload_notebook(
                client,
                superuser_token_headers,
                filename="stateful_simple.ipynb",
            )
            data = await wait_for_event(websocket, "notebooks", tries=100)
            assert len(data["notebooks"]) == 2
            assert data["notebooks"][0]["uuid"] == uuid
            assert data["notebooks"][0]["name"] == "simple"
            assert data["notebooks"][1]["uuid"] == uuid2
            assert data["notebooks"][1]["name"] == "stateful_simple"

            response = await client.delete(
                f"/notebooks/{uuid}", headers=superuser_token_headers
            )
            assert response.status_code == 204
            data = await wait_for_event(websocket, "notebooks", tries=100)
            assert len(data["notebooks"]) == 1
            assert data["notebooks"][0]["uuid"] == uuid2
            assert data["notebooks"][0]["name"] == "stateful_simple"

            response = await client.delete(
                "/notebooks", headers=superuser_token_headers
            )
            assert response.status_code == 204
            data = await wait_for_event(websocket, "notebooks", tries=100)
            assert len(data["notebooks"]) == 0



from pprint import pformat

from asyncblink import signal  # type: ignore
from fastapi.logger import logger

from jupyter_d1.signals import (
    CONTROL_CHANNEL,
    HB_CHANNEL,
    IOPUB_CHANNEL,
    SHELL_CHANNEL,
    STDIN_CHANNEL,
)

log_to_file = False
log_to_logger = False


class KernelLogger:
    def __init__(self):
        signal(IOPUB_CHANNEL).connect(self.log_iopub)
        signal(SHELL_CHANNEL).connect(self.log_shell)
        signal(STDIN_CHANNEL).connect(self.log_stdin)
        signal(HB_CHANNEL).connect(self.log_hb)
        signal(CONTROL_CHANNEL).connect(self.log_control)
        if log_to_file:
            self.file = open("/tmp/kernel_msg.log", "w")

    def log(self, channel, msg):
        if log_to_file:
            self.file.write(
                f"\n>>>>>>>>>>>>>>>> kernel >>>>>>>>>>>>>>>>>>>>\n"
            )
            self.file.write(f">>> {channel}: {pformat(msg)}\n")
        if log_to_logger:
            logger.debug(f"{channel}: {pformat(msg)}")

    async def log_iopub(self, sender, msg, **kwargs):
        self.log("iopub", msg)

    async def log_shell(self, sender, msg, **kwargs):
        self.log("shell", msg)

    async def log_stdin(self, sender, msg, **kwargs):
        self.log("stdin", msg)

    async def log_hb(self, sender, msg, **kwargs):
        self.log("hb", msg)

    async def log_control(self, sender, msg, **kwargs):
        self.log("control", msg)



from ..kernel_definition import KernelDefinition
from .vars_manager import VarsManager, ZshVarsManager
from .workdir_manager import WorkDirManager, ZshWorkDirManager


class ZshKernelDefinition(KernelDefinition):
    def create_vars_manager(self) -> VarsManager:
        return ZshVarsManager()

    def create_workdir_manager(self, workdir: str) -> WorkDirManager:
        return ZshWorkDirManager(workdir)


kernel_definition = ZshKernelDefinition()



from typing import Optional

from . import python, r, zsh
from .kernel_definition import KernelDefinition
from .vars_manager import VarsManager  # noqa
from .workdir_manager import WorkDirManager  # noqa


def get_kernel_definition(language: str) -> Optional[KernelDefinition]:
    if language == "python":
        return python.kernel_definition
    elif language in ("zsh", "bash"):
        return zsh.kernel_definition
    elif language == "R":
        return r.kernel_definition
    else:
        return None



import asyncio
from typing import Any

import requests
from asyncblink import signal  # type: ignore
from fastapi import APIRouter, Depends, Request

from jupyter_d1.signals import RCLONE_UPDATE

from ..d1_response import D1Response
from ..deps import write_access
from ..settings import settings

router = APIRouter(default_response_class=D1Response)


@router.post("/{api_path:path}", dependencies=[Depends(write_access)])
async def rclone_api(api_path: str, request: Request) -> Any:
    body = await request.body()
    r = requests.post(
        f"http://127.0.0.1:5572/{api_path}",
        params=request.query_params,
        headers=request.headers,
        data=body,
    )
    return r.json()


def dispatch_rclone_stats():
    r = None
    try:
        r = requests.post(f"http://127.0.0.1:5572/core/stats", timeout=3)
    except Exception:
        pass
    if r is not None:
        signal(RCLONE_UPDATE).send(stats=r.json())


async def stats_periodic():
    while True:
        dispatch_rclone_stats()
        await asyncio.sleep(settings.RCLONE_STATS_POLLING_INTERVAL)



from typing import List, Optional

from fastapi.logger import logger

from ...models.kernel_variable import KernelVariable
from ...utils import NotebookNode
from ..vars_manager import VarsManager

raw_get_vars_code = """
env
"""  # noqa


class ZshVarsManager(VarsManager):
    def get_vars_code(self) -> str:
        return raw_get_vars_code

    def get_single_var_code(self, var_name: str) -> str:
        return ""

    def parse_vars_response(
        self, vars_response: NotebookNode
    ) -> List[KernelVariable]:
        vars = []
        try:
            line = vars_response.text.strip()
            name, value = line.split("=", 1)

            vars.append(
                KernelVariable(
                    name=name,
                    type=None,
                    value={"single_value": value},
                    summary=value,
                )
            )
        except Exception as e:
            logger.debug(
                f"Exception parsing var for zsh kernel: {e}, "
                f"{vars_response.text}"
            )
        return vars

    def parse_single_var_response(
        self, var_response: NotebookNode
    ) -> Optional[KernelVariable]:
        return None



import asyncio
from datetime import datetime
from typing import AsyncGenerator, Dict

import pytest  # type: ignore
from async_asgi_testclient import TestClient  # type: ignore
from pytest_mock import MockFixture  # type: ignore

from jupyter_d1 import app
from jupyter_d1.models.server_stats import (
    CPUStats,
    DiskStats,
    GPUStats,
    MemoryStats,
    ServerStats,
)
from jupyter_d1.settings import settings

from .utils import wait_for_event


@pytest.fixture()
async def client() -> AsyncGenerator:
    async with TestClient(app) as c:
        yield c


@pytest.fixture()
async def superuser_token_headers(client: TestClient) -> Dict[str, str]:
    r = await client.get(
        "/login/access-token", headers={"Authorization": "test9token_4"}
    )
    token = r.json()["token"]["access_token"]
    headers = {"Authorization": f"Bearer {token}"}
    return headers


class TestStats:
    @pytest.mark.asyncio
    async def test_get_stats(
        self, client: TestClient, superuser_token_headers: Dict[str, str],
    ):
        await asyncio.sleep(1)
        response = await client.get(
            f"/server/stats", headers=superuser_token_headers
        )
        assert response.status_code == 200
        stats = response.json()["server_stats"]
        assert len(stats) >= 1
        assert stats[0]["timestamp"] is not None
        assert stats[0]["disk"]["free"] is not None
        assert stats[0]["disk"]["percent"] is not None
        assert stats[0]["memory"]["available"] is not None
        assert stats[0]["memory"]["total"] is not None
        assert stats[0]["cpu"]["percent"] is not None
        assert stats[0]["cpu"]["load_percent"] is not None
        assert stats[0]["gpu"] == []

    @pytest.mark.asyncio
    async def test_get_latest_stats(
        self, client: TestClient, superuser_token_headers: Dict[str, str],
    ):
        await asyncio.sleep(1)
        response = await client.get(
            f"/server/latest_stats", headers=superuser_token_headers
        )
        assert response.status_code == 200
        stats = response.json()["server_stats"]
        assert stats["timestamp"] is not None
        assert stats["disk"]["free"] is not None
        assert stats["disk"]["percent"] is not None
        assert stats["memory"]["available"] is not None
        assert stats["memory"]["total"] is not None
        assert stats["cpu"]["percent"] is not None
        assert stats["cpu"]["load_percent"] is not None
        assert stats["gpu"] == []

    @pytest.mark.asyncio
    async def test_stats_response(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        mocker: MockFixture,
    ):
        stats_manager = mocker.patch("jupyter_d1.routers.server.stats_manager")
        stats_manager.get_stats.return_value = [
            ServerStats(
                timestamp=datetime(1994, 5, 3, 4, 3, 4),
                disk=DiskStats(free=5, percent=9.3, total=300),
                memory=MemoryStats(available=400, total=953),
                cpu=CPUStats(percent=8.73, load_percent=44.3),
                gpu=[
                    GPUStats(
                        name="rrr30 d",
                        gpu_usage_percent=44.43,
                        memory_usage_percent=8.1,
                        memory_free=4420,
                        memory_used=999,
                        memory_total=1009,
                    )
                ],
            )
        ]
        response = await client.get(
            f"/server/stats", headers=superuser_token_headers
        )
        assert response.status_code == 200
        stats = response.json()["server_stats"]
        assert len(stats) == 1
        assert stats[0]["timestamp"] == "1994-05-03T04:03:04"
        assert stats[0]["disk"]["free"] == 5
        assert stats[0]["disk"]["percent"] == 9.3
        assert stats[0]["disk"]["total"] == 300
        assert stats[0]["memory"]["available"] == 400
        assert stats[0]["memory"]["total"] == 953
        assert stats[0]["cpu"]["percent"] == 8.73
        assert stats[0]["cpu"]["load_percent"] == 44.3
        assert len(stats[0]["gpu"]) == 1
        assert stats[0]["gpu"][0]["name"] == "rrr30 d"
        assert stats[0]["gpu"][0]["gpu_usage_percent"] == 44.43
        assert stats[0]["gpu"][0]["memory_usage_percent"] == 8.1
        assert stats[0]["gpu"][0]["memory_free"] == 4420
        assert stats[0]["gpu"][0]["memory_used"] == 999
        assert stats[0]["gpu"][0]["memory_total"] == 1009

    @pytest.mark.asyncio
    async def test_get_environment(
        self, client: TestClient, superuser_token_headers: Dict[str, str],
    ):
        await asyncio.sleep(1)
        response = await client.get(
            f"/server/environment", headers=superuser_token_headers
        )
        assert response.status_code == 200
        environment = response.json()["environment_info"]
        assert environment["os"]["platform"] is not None
        assert environment["os"]["name"] is not None
        assert environment["os"]["raw_name"] is not None
        assert environment["os"]["uname"] is not None
        assert environment["python"]["version"] is not None
        assert environment["python"]["executable"] is not None
        assert environment["python"]["pythonpath"] is not None
        assert environment["python"]["version_info"] is not None
        assert environment["python"]["packages"] is not None
        assert environment["python"]["version_info"]["major"] is not None
        assert environment["python"]["version_info"]["minor"] is not None
        assert environment["python"]["version_info"]["micro"] is not None
        assert (
            environment["python"]["version_info"]["releaselevel"] is not None
        )
        assert environment["python"]["version_info"]["serial"] is not None
        assert environment["process"]["argv"] is not None
        assert environment["process"]["cwd"] is not None
        assert environment["process"]["user"] is not None
        assert environment["process"]["pid"] is not None
        assert environment["process"]["environ"] is not None
        assert environment["config"] is not None
        if environment.get("conda", None) is not None:
            assert environment["conda"]["version"] is not None
        assert environment["cores"] is not None
        assert environment["version"] == settings.VERSION

    @pytest.mark.asyncio
    async def test_get_version(
        self, client: TestClient, superuser_token_headers: Dict[str, str],
    ):
        response = await client.get(
            f"/server/version", headers=superuser_token_headers
        )
        assert response.json()["version"] == settings.VERSION


class TestStatsWebSocket:
    @pytest.mark.asyncio
    async def test_stats_websocket(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):

        async with client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:

            stats = await wait_for_event(websocket, "server_stats")
            assert stats["server_stats"]["timestamp"] is not None
            assert stats["server_stats"]["disk"]["free"] is not None
            assert stats["server_stats"]["disk"]["percent"] is not None
            assert stats["server_stats"]["disk"]["total"] is not None
            assert stats["server_stats"]["memory"]["available"] is not None
            assert stats["server_stats"]["memory"]["total"] is not None
            assert stats["server_stats"]["cpu"]["percent"] is not None
            assert stats["server_stats"]["cpu"]["load_percent"] is not None
            assert stats["server_stats"]["gpu"] == []


class TestLogWebSocket:
    @pytest.mark.asyncio
    async def test_log_websocket(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):

        async with client.websocket_connect(
            f"/server/ws", headers=superuser_token_headers
        ) as websocket:

            from jupyter_d1.main import logger

            logger.debug("hi hello jjkolq")

            log = await wait_for_event(websocket, "log")
            assert "hi hello jjkolq" in log["log"]
            assert "DEBUG" in log["log"]



import asyncio
from typing import Dict, List
from uuid import UUID

from asyncblink import signal  # type: ignore
from fastapi.logger import logger
from watchdog.observers.api import ObservedWatch  # type: ignore
from wsgidav.wsgidav_app import WsgiDAVApp  # type: ignore

from .settings import settings
from .signals import WATCHDOG_FILE_SYSTEM_EVENT

try:
    if not settings.WATCHDOG_ENABLED:
        raise ImportError("settings.WATCHDOG_ENABLED is False")
    from watchdog.events import FileSystemEvent  # type: ignore
    from watchdog.events import PatternMatchingEventHandler
    from watchdog.observers import Observer  # type: ignore

    has_watchdog = True
    logger.debug("Successfully loaded watchdog.")
except ImportError as e:
    has_watchdog = False
    logger.debug(f"Watchdog disabled.  Failed to import - {e}")


if has_watchdog:

    async def send_watchdog_event(event: FileSystemEvent):
        signal(WATCHDOG_FILE_SYSTEM_EVENT).send(
            event_type=event.event_type,
            src_path=event.src_path,
            dest_path=getattr(event, "dest_path", None),
        )

    class D1FileSystemEventHandler(PatternMatchingEventHandler):
        def __init__(self, loop, *args, **kwargs):
            self.loop = loop
            super().__init__(*args, **kwargs)

        def on_any_event(self, event: FileSystemEvent):
            asyncio.run_coroutine_threadsafe(
                send_watchdog_event(event), self.loop,
            )

    class WatchdogManager:
        def __init__(self, loop):
            self.watches: Dict[str, ObservedWatch] = {}
            self.observer = Observer()
            self.event_handler = D1FileSystemEventHandler(
                loop, ignore_patterns=settings.WATCHDOG_IGNORE_PATTERNS
            )

        def start_watchdog(self):
            if not self.observer.is_alive():
                self.observer.schedule(
                    self.event_handler,
                    settings.ROOT_DIR,
                    recursive=True,
                    exclude_dirs=[
                        bytes(p, "utf-8") for p in settings.WATCHDOG_IGNORE_DIRS
                    ]
                )
                self.observer.start()

        def stop_watchdog(self):
            self.observer.stop()

        def schedule(self, directory: str, exclude_dirs: List[str] = []):
            watch = self.observer.schedule(
                self.event_handler,
                directory,
                recursive=True,
                exclude_dirs=[bytes(p, "utf-8") for p in exclude_dirs],
            )
            self.watches[directory] = watch

        def unschedule(self, directory: str):
            watch = self.watches.pop(directory, None)
            if watch:
                self.observer.unschedule(watch)

    def create_watchdog(loop):
        return WatchdogManager(loop)


else:

    def create_watchdog(loop):
        return None


class D1WsgiDAVApp(WsgiDAVApp):
    def remove_providers(self, shares: List[str]):
        for share in shares:
            self.provider_map.pop(share, None)

        # Store the list of share paths, ordered by length, so route lookups
        # will return the most specific match
        self.sorted_share_list = [s.lower() for s in self.provider_map.keys()]
        self.sorted_share_list = sorted(
            self.sorted_share_list, key=len, reverse=True
        )


dav = D1WsgiDAVApp(settings.DAV_SETTINGS)


def format_share(directory: str) -> str:
    return "/" + directory.lower().strip("/")


class DAVManager:
    def __init__(self):
        # Dict of shares used by notebooks (excluding the D1 working
        # directory and any default shares that wsgi_dav provides).
        # A share is just a route exposed by wsgi_dav,
        # i.e. "/tmp/docs" is the share for the /tmp/docs directory, where
        # https://host/dav/tmp/docs is its dav url
        self.nb_to_directory: Dict[UUID, str] = {}
        self.root_share = format_share(settings.ROOT_DIR)
        self.watchdog_manager = None

    def start_watchdog(self, loop):
        if self.watchdog_manager is None:
            watchdog_manager = create_watchdog(loop)
            if watchdog_manager:
                watchdog_manager.start_watchdog()
            self.watchdog_manager = watchdog_manager

    def stop_watchdog(self):
        if self.watchdog_manager:
            self.watchdog_manager.stop_watchdog()
            self.watchdog_manager = None

    def add_provider(self, directory: str, uuid: UUID) -> None:
        self.nb_to_directory[uuid] = directory
        self.update_providers()

    def remove_provider(self, uuid: UUID) -> None:
        self.nb_to_directory.pop(uuid, None)
        self.update_providers()

    def clear_providers(self) -> None:
        self.nb_to_directory = {}
        self.update_providers()

    def update_providers(self):
        """
        Iterate through working directories for all open notebooks in order
        of length and find the resolved dav providers. If the provider is
        for a less specific directory the first time a provider is resolved
        to, then the provider should be replaced with a more specific
        provider. If a provider is never resolved to, remove it.
        If a provider is not found for a directory, add a provider for that
        directory. The nb_to_directory dict should reflect the desired state
        before this method is called.
        """

        # Order directories by ascending length so that shares that are too
        # general will be caught first
        directories = [s.lower() for s in self.nb_to_directory.values()]
        directories = sorted(directories, key=len, reverse=False)
        used_shares = set()
        unused_shares = set()
        add_shares = set()

        for directory in directories:
            # Check if an existing provider contains the required directory
            dav_share, _ = dav.resolve_provider(directory.lower())
            if dav_share is not None:
                # if the share is the root dir share, do nothing
                if dav_share == self.root_share:
                    continue
                if dav_share not in used_shares and len(dav_share) < len(
                    directory
                ):
                    # share needs to be more specific, so mark it
                    # for replacement
                    unused_shares.add(dav_share)
                    add_shares.add(directory)
                else:
                    # share found, add to list of used shares
                    used_shares.add(dav_share)
            else:
                # share not found, add to list of provider to add
                add_shares.add(directory)

        # Get unused shares
        for share in dav.sorted_share_list:
            if share == self.root_share:
                continue
            if share not in used_shares:
                unused_shares.add(share)
            # check if any shares being added can replace an old share
            for add_share in add_shares:
                if share.lower().startswith(add_share + "/"):
                    unused_shares.add(share)

        if len(unused_shares) > 0:
            dav.remove_providers(unused_shares)
        if self.watchdog_manager:
            for share in unused_shares:
                self.watchdog_manager.unschedule(share)

        for directory in add_shares:
            formatted_share = format_share(directory)
            dav.add_provider(formatted_share, directory)
            if self.watchdog_manager:
                self.watchdog_manager.schedule(directory)


dav_manager = DAVManager()



# Standard names for Blinker signals shared between modules

# Signals that are fired when new data comes in on respective channels.
# The signal will be emitted with these kwargs:
#    * msg - contents of the message
#    * kernel_id - uuid of the kernel emitting the message
#    * channel - name of the channel, eg iopub, shell, etc
IOPUB_CHANNEL = "iopub_channel"
SHELL_CHANNEL = "shell_channel"
STDIN_CHANNEL = "stdin_channel"
HB_CHANNEL = "hb_channel"
CONTROL_CHANNEL = "control_channel"

# Signal fired when a cell in a document changes
# The signal will be emitted with these kwargs:
#    * cell - The nbformat.notebooknode.NotebookNode object that is the cell
#    * notebook_id - The uuid of the notebook that the cell belongs to
CELL_UPDATE = "cell_update"
CELL_ADDED = "cell_added"
CELL_DELETED = "cell_deleted"

# Signal fired when a cell produces stream output
# The signal will be emitted with these kwargs:
#    * output - The nbformat-style cell output
#    * notebook_id - The uuid of the notebook that the cell belongs to
#    * cell_id - The uuid of the cell
CELL_STREAM = "cell_stream"

# Signal fired when a cell execution_count gets updated
# The signal will be emitted with these kwargs:
#    * execution_count - the int execution count for the cell
#    * notebook_id - The uuid of the notebook that the cell belongs to
#    * cell_id - The uuid of the cell
#    * parent_id - The message_id of the original execution request
CELL_EXECUTION_REPLY = "cell_execution_reply"

# Signal fired when the kernel starts to execute a piece of code
# The signal will be emitted with these kwargs:
#    * content - the content dictionary from the original message including
#       * code - the code being executed
#       * execution_count - the int execution count for the cell
#    * notebook_id - The uuid of the notebook that the cell belongs to
#    * parent_id - The message_id of the original execution request
CELL_EXECUTION_INPUT = "cell_execution_input"

# Signal fired when the kernel starts to execute a piece of code
# The signal will be emitted with these kwargs:
#    * notebook_id - The uuid of the notebook that the cell belongs to
#    * cell_id - The uuid of the cell
#    * message_id - The message_id of the  execution request
CELL_EXECUTION_REQUEST = "cell_execution_request"

# Signal fired when a scratch command updates. This may happen multiple times
# for a single scratch command with different outputs.
# The signal will be emitted with these kwargs:
#    * msg_id - The msg_id of the scratch command execution
#    * notebook_id - The uuid of the notebook that the scratch command was
#                    run in
#    * output - Output of the scratch command
#    * execution_state - The state of the scratch command
SCRATCH_UPDATE = "scratch_update"

# Signal fired when a kernels vars are updated
# The signal will be emitted with these kwargs:
#    * notebook_id - The uuid of the notebook
#    * vars - List of KernelVariable objects
VARS_UPDATE = "vars_update"

# Signal fired when a details response for a variable comes back
# The signal will be emitted with these kwargs:
#    * notebook_id - The uuid of the notebook
#    * single_var - Optional KernelVariable object
VAR_DETAILS = "var_details"

# Signal fired when the kernel emits a page payload (like help command
# output).
# The signal will be emitted with these kwargs:
#    * msg_id - The msg_id of the command that prompted this payload
#    * notebook_id - The uuid of the notebook that the command was
#                    run in
#    * data - The mimebundle of data, includes at least text/plain
#    * start - Line offset to start from
PAYLOAD_PAGE = "payload_page"

# Signal fired when code complete results come back
# The signal will be emitted with these kwargs:
#    * msg_id - The msg_id of the completion request
#    * notebook_id - The uuid of the notebook
#    * matches - List of possible completions
#    * cursor_start - Start of text that would be replaced
#    * cursor_end - End of text that would be replaced
COMPLETE_REPLY = "complete_reply"

HISTORY_REPLY = "history_reply"
METADATA_UPDATE = "metadata_update"
KERNEL_RESTARTED = "kernel_restarted"
KERNEL_INTERRUPTED = "kernel_interrupted"

# Signals fired when notebook is added to or removed from notebook manager
# The signal will be emitted with these kwargs:
#    * notebooks - The dict of notebook uuid to notebook nodes from the
#                  notebook manager
NOTEBOOKS_UPDATED = "notebook_updated"

# FastAPI lifecycle events
APP_STARTUP = "app_startup"
APP_SHUTDOWN = "app_shutdown"

RCLONE_UPDATE = "rclone_update"
WATCHDOG_FILE_SYSTEM_EVENT = "watchdog_event"
SERVER_STATS_UPDATE = "server_stats_update"
APP_LOG = "app_log"



import json
from typing import List, Optional

from fastapi.logger import logger

from ...models.kernel_variable import KernelVariable
from ...settings import settings
from ...utils import NotebookNode
from ..vars_manager import VarsManager

raw_install_callistor_github_code = """
        if (!("remotes" %in% (.packages()))) {{
            if (length(find.package("remotes", quiet=TRUE)) < 1) {{
                install.packages("remotes")
            }}
        }}
        remotes::install_github("OakCityLabs/callisto-r@{branch}", upgrade="never", quiet=TRUE)
"""  # noqa

raw_install_callistor_cran_code = """
        install.packages("callistor")
"""  # noqa

raw_load_callistor_code = """
if (!("callistor" %in% (.packages()))) {{
    if (length(find.package("callistor", quiet=TRUE)) < 1) {{
        {install_callistor_code}
    }}
    library("callistor")
}}
"""  # noqa

raw_get_vars_code = """
{load_callistor_code}

cat(callistor::format_vars(environment(), {abbrev_len}))
"""  # noqa

# Note: we try to access the var first so it throws if it doesn't exist
raw_get_single_var_code = """
{load_callistor_code}

cat(callistor::format_var(environment(), "{var_name}", NULL))
"""  # noqa


def get_load_callistor_code(from_github=True) -> str:
    return raw_load_callistor_code.format(
        install_callistor_code=raw_install_callistor_github_code.format(
            branch=settings.CALLISTOR_GITHUB_BRANCH
        )
        if from_github
        else raw_install_callistor_cran_code
    )


class RVarsManager(VarsManager):
    def get_vars_code(self) -> str:
        return raw_get_vars_code.format(
            load_callistor_code=get_load_callistor_code(
                settings.LOAD_CALLISTOR_FROM_GITHUB
            ),
            abbrev_len=settings.VAR_ABBREV_LEN,
        )

    def get_single_var_code(self, var_name: str) -> str:
        return raw_get_single_var_code.format(
            load_callistor_code=get_load_callistor_code(
                settings.LOAD_CALLISTOR_FROM_GITHUB
            ),
            var_name=var_name,
        )

    def parse_vars_response(
        self, vars_response: NotebookNode
    ) -> List[KernelVariable]:
        vars: List[KernelVariable] = []
        if "text" not in vars_response:
            return vars
        try:
            json_vars = json.loads(vars_response.text)
            for json_var in json_vars:
                vars.append(
                    KernelVariable(
                        name=json_var.get("name"),
                        type=json_var.get("type"),
                        abbreviated=json_var.get("abbreviated"),
                        summary=str(json_var.get("summary")),
                        value=json_var.get("value"),
                    )
                )
        except Exception as e:
            print(e)
            logger.debug(
                f"Exception parsing vars for R kernel: {e}, "
                f"{vars_response.text}"
            )
        return vars

    def parse_single_var_response(
        self, var_response: NotebookNode
    ) -> Optional[KernelVariable]:
        var = None
        if "text" not in var_response:
            return var
        try:
            json_var = json.loads(var_response.text)
            var = KernelVariable(
                name=json_var.get("name"),
                type=json_var.get("type"),
                abbreviated=json_var.get("abbreviated"),
                value=json_var.get("value"),
                summary=json_var.get("summary"),
            )
        except Exception as e:
            logger.debug(
                f"Exception parsing var for R kernel: {e}, "
                f"{var_response.text}"
            )
        return var



from pathlib import Path
from typing import List

from ..kernel_definition import KernelDefinition
from .vars_manager import PythonVarsManager, VarsManager
from .workdir_manager import PythonWorkDirManager, WorkDirManager


class PythonKernelDefinition(KernelDefinition):
    def create_vars_manager(self) -> VarsManager:
        return PythonVarsManager()

    def create_workdir_manager(self, workdir: str) -> WorkDirManager:
        return PythonWorkDirManager(workdir)

    @property
    def kernel_options(self) -> List[str]:
        return [
            f"--IPKernelApp.exec_files=['{Path(__file__).parent.resolve()}"
            f"/startup/callisto_startup.py']"
        ]


kernel_definition = PythonKernelDefinition()



import json
from enum import Enum

import requests

from jupyter_d1.settings import settings


class D1CommandType(str, Enum):
    NOTIFY = "notify"


def execute_d1_notify(title: str, message: str):
    requests.post(
        f"{settings.MOTHERSHIP_URL}/work_nodes/{settings.WORK_NODE_ID}/push",
        json={
            "secret": settings.PUSH_NOTE_SECRET_KEY,
            "title": title,
            "message": message,
        },
        timeout=30,
    )


def execute_d1_command(command: str):
    try:
        command_dict = json.loads(command)
        command_type = D1CommandType(command_dict.get("command_type"))
        if command_type == D1CommandType.NOTIFY:
            execute_d1_notify(command_dict["title"], command_dict["message"])
    except Exception:
        pass



from typing import Dict

from fastapi.testclient import TestClient


def test_get_access_token(client: TestClient) -> None:
    r = client.get(f"/login/access-token", headers={"Authorization": "bogus"})
    assert r.status_code == 401

    r = client.get(f"/login/access-token")
    assert r.status_code == 401

    r = client.get(
        f"/login/access-token", headers={"Authorization": "test9token_4"}
    )
    tokens = r.json()["token"]
    assert r.status_code == 200
    assert "access_token" in tokens
    assert tokens["access_token"]


def test_use_access_token(
    client: TestClient, superuser_token_headers: Dict[str, str]
) -> None:
    r = client.post(f"/login/test-token", headers=superuser_token_headers,)
    result = r.json()["permission"]
    assert r.status_code == 200
    assert result["id"] == -1
    assert result["user_id"] == -1
    assert result["work_node_id"] == -1
    assert result["read_access"] is True
    assert result["write_access"] is True



from collections import deque
from datetime import datetime
from threading import RLock
from typing import Deque, List, Optional

from asyncblink import signal  # type: ignore
from fastapi.logger import logger

try:
    import psutil  # type: ignore

    has_psutil = True
    logger.debug("Successfully loaded psutil.")
except ImportError:
    has_psutil = False
    logger.debug("Failed to load psutil.  Disabling.")
import asyncio
from itertools import islice

try:
    from pynvml import NVMLError  # type: ignore
    from pynvml import (
        nvmlDeviceGetCount,
        nvmlDeviceGetHandleByIndex,
        nvmlDeviceGetMemoryInfo,
        nvmlDeviceGetName,
        nvmlDeviceGetUtilizationRates,
        nvmlInit,
    )

    nvml_support = True
except ImportError:
    nvml_support = False

from ..models.server_stats import (
    CPUStats,
    DiskStats,
    GPUStats,
    MemoryStats,
    ServerStats,
)
from ..settings import settings
from ..signals import SERVER_STATS_UPDATE

if nvml_support:
    try:
        nvmlInit()
        device_count = nvmlDeviceGetCount()
    except NVMLError as err:
        print(f"Error with init NVML: {err}")
        nvml_support = False


class StatsManager:
    def __init__(self, max_cache_length: int):
        self.max_cache_length = max_cache_length
        self.cache: Deque[ServerStats] = deque(maxlen=max_cache_length)
        self.cache_lock = RLock()

    def get_stats(self, start, end) -> List[ServerStats]:
        return list(islice(self.cache, start, end))

    def get_latest(self) -> Optional[ServerStats]:
        return self.cache[0] if len(self.cache) > 0 else None

    def put(self, stats: ServerStats):
        with self.cache_lock:
            self.cache.appendleft(stats)

    def get_gpu_stats(self) -> List[GPUStats]:
        if not nvml_support or device_count == 0:
            return []
        gpu_stats_list: List[GPUStats] = []
        for i in range(device_count):
            handle = nvmlDeviceGetHandleByIndex(i)

            gpu_stats = GPUStats()
            try:
                gpu_stats.name = nvmlDeviceGetName(handle)
                util = nvmlDeviceGetUtilizationRates(handle)
                gpu_stats.gpu_usage_percent = util.gpu
                gpu_stats.memory_usage_percent = util.memory
                memory_info = nvmlDeviceGetMemoryInfo(handle)
                gpu_stats.memory_free = memory_info.free
                gpu_stats.memory_used = memory_info.used
                gpu_stats.memory_total = memory_info.total
            except NVMLError as err:
                print(f"Error getting gpu stats: {err}")
            gpu_stats_list.append(gpu_stats)
        return gpu_stats_list

    def collect_server_stats(self) -> ServerStats:
        if has_psutil:
            disk_usage = psutil.disk_usage(settings.ROOT_DIR)
            memory_usage = psutil.virtual_memory()
            cpu_usage = psutil.cpu_percent()
            cpu_load_percent = psutil.getloadavg()[0] / psutil.cpu_count()
        else:
            disk_usage = 0
            memory_usage = 0
            cpu_usage = 0
            cpu_load_percent = 0
        gpu_stat_list = self.get_gpu_stats()
        server_stats = ServerStats(
            timestamp=datetime.now(),
            cpu=CPUStats(percent=cpu_usage, load_percent=cpu_load_percent),
            disk=DiskStats(
                free=disk_usage.free,
                percent=disk_usage.percent,
                total=disk_usage.total,
            ),
            memory=MemoryStats(
                available=memory_usage.available, total=memory_usage.total
            ),
            gpu=gpu_stat_list,
        )
        self.put(server_stats)
        return server_stats

    def dispatch_server_stats(self):
        server_stats = None
        try:
            server_stats = self.collect_server_stats()
        except Exception:
            pass
        if server_stats is not None:
            signal(SERVER_STATS_UPDATE).send(server_stats=server_stats)

    async def server_stats_periodic(self):
        while True:
            self.dispatch_server_stats()
            await asyncio.sleep(settings.SERVER_STATS_POLLING_INTERVAL)



import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile
import traceback

from fastapi import APIRouter, File, HTTPException, UploadFile, status
from fastapi.logger import logger
from jupytext.cli import jupytext  # type: ignore
from starlette.responses import StreamingResponse

from ..d1_response import D1Response

COPY_BUFSIZE = 64 * 1024

router = APIRouter(default_response_class=D1Response)


@router.post(
    "/rmd_to_ipynb", status_code=status.HTTP_200_OK,
)
async def convert_rmd_to_ipynb(file: UploadFile = File(...)):

    out_file = NamedTemporaryFile(mode="rb", suffix=".ipynb")
    try:

        with NamedTemporaryFile(mode="wb", suffix=".Rmd") as f:
            shutil.copyfileobj(file.file, f)
            f.seek(0)
            jupytext(
                [
                    f.name,
                    "--from",
                    "rmarkdown",
                    "--to",
                    "notebook",
                    "--output",
                    out_file.name,
                    "--set-kernel",
                    "ir",
                ]
            )
    except Exception as e:
        logger.info(
            f"Failed to convert Rmd file to ipynb {e} \n"
            f"{traceback.format_exc()}"
        )
        msg = "Failed to convert Rmd file to ipynb"
        if isinstance(e, KeyError):
            msg += ", do you have an R kernel installed and added to the kernelspecs?"
        raise HTTPException(400, msg)

    filename = file.filename
    if filename is None or len(filename) < 1:
        filename = "notebook"
    headers = {
        "content-disposition": f'attachment; filename="{Path(filename).stem}.ipynb"'
    }
    out_file.seek(0)

    def iterfile():
        yield from out_file

    return StreamingResponse(
        content=iterfile(),
        headers=headers,
        media_type="application/x-ipynb+json",
    )
